﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;
using PseudoInterpreter.Ast;

namespace PseudoInterpreter.UI
{
    /// <summary>
    /// Logica di interazione per AstControl.xaml
    /// </summary>
    public partial class AstControl : UserControl
    {
        public AstControl()
        {
            InitializeComponent();
            try
            {
                AstParser parser = new AstParser(new System.IO.StringReader(CodeProvider.Code));
                ProgramStatement root = parser.ParseTree();
                var items = CreateNodes(root, true, "");
                foreach (var i in items)
                {
                    AstTree.Items.Add(i);
                }
            }
            catch (Exception ex) {
                this.Foreground = new SolidColorBrush(Color.FromRgb(0xdc, 0x32, 0x2f));
                this.Content = ex.Message;
            }
        }
        private static TreeViewItem[] CreateNodes(Node node, bool content, string prefix) {
            if (node == null) { return new TreeViewItem[]{ new TreeViewItem(){ Header=prefix + "[Empty]" } }; }
            switch (node.NodeType) {
                case NodeType.ArrayExpression:
                    return CreateArrayAccessor((ArrayExpression)node, content, prefix);
                case NodeType.ArrayRangeExpression:
                    return CreateArrayRange((ArrayRangeExpression)node, content, prefix);
                case NodeType.BinaryExpression:
                    return CreateBinary((BinaryExpression)node, content, prefix);
                case NodeType.BlockStatement:
                    return CreateBlock((BlockStatement)node, content, prefix);
                case NodeType.ConstantFloat:
                case NodeType.ConstantInfinity:
                case NodeType.ConstantInteger:
                case NodeType.ConstantNil:
                case NodeType.ConstantEmptySet:
                    return CreateConstant((Constant)node, content, prefix);
                case NodeType.ExpressionStatement:
                    return CreateExpression((ExpressionStatement)node, content, prefix);
                case NodeType.ForStatement:
                    return CreateFor((ForStatement)node, content, prefix);
                case NodeType.ForEachStatement:
                    return CreateForEach((ForEachStatement)node, content, prefix);
                case NodeType.CallExpression:
                    return CreateFunctionCall((CallExpression)node, content, prefix);
                case NodeType.FunctionDeclaration:
                    return CreateFunctionDefinition((FunctionDeclaration)node, content, prefix);
                case NodeType.GroupExpression:
                    return CreateGroup((GroupExpression)node, content, prefix);
                case NodeType.Identifier:
                    return CreateIdentifier((Identifier)node, content, prefix);
                case NodeType.IfStatement:
                    return CreateIf((IfStatement)node, content, prefix);
                case NodeType.MemberAccessor:
                    return CreateMemberAccessor((MemberAccessor)node, content, prefix);
                case NodeType.ProgramStatement:
                    return CreateProgram((ProgramStatement)node, content, prefix);
                case NodeType.ReturnStatement:
                    return CreateReturn((ReturnStatement)node, content, prefix);
                case NodeType.UnaryExpression:
                    return CreateUnary((UnaryExpression)node, content, prefix);
                case NodeType.WhileStatement:
                    return CreateWhile((WhileStatement)node, content, prefix);
                case NodeType.RepeatStatement:
                    return CreateRepeat((RepeatStatement)node, content, prefix);
                case NodeType.PrintStatement:
                    return CreatePrint((PrintStatement)node, content, prefix);
                case NodeType.SwapStatement:
                    return CreateSwap((SwapStatement)node, content, prefix);
                case NodeType.ArrayInitializationExpression:
                    return CreateArrayInitializationExpression((ArrayInitializationExpression)node, content, prefix);
                case NodeType.SetInitializationExpression:
                    return CreateSetInitializationExpression((SetInitializationExpression)node, content, prefix);
            }
            return null;
        }

        private static TreeViewItem[] CreateIdentifier(Identifier i, bool content, string prefix) {
            if (content) { return new TreeViewItem[] { }; }
            return new TreeViewItem[] {
                new TreeViewItem(){ Header = prefix + "[Identifier " + i.Name + "]" }
            };
        }
        private static TreeViewItem[] CreateConstant(Constant c, bool content, string prefix) {
            if (content) { return new TreeViewItem[] { }; }
            string txt = "";
            switch (c.NodeType)
            {
                case NodeType.ConstantFloat:
                    txt = prefix + "[ConstantDouble " + ((ConstantFloat)c).Value + "]";
                    break;
                case NodeType.ConstantInfinity:
                    txt = prefix + "[ConstantInfinity]";
                    break;
                case NodeType.ConstantInteger:
                    txt = prefix + "[ConstantInteger " + ((ConstantInteger)c).Value + "]";
                    break;
                case NodeType.ConstantNil:
                    txt = prefix + "[ConstantNil]";
                    break;
                case NodeType.ConstantEmptySet:
                    txt = prefix + "[ConstantEmptySet]";
                    break;
            }
            return new TreeViewItem[] { new TreeViewItem() { Header = txt } };
        }
        private static TreeViewItem[] CreateArrayInitializationExpression(ArrayInitializationExpression e, bool content, string prefix)
        {
            if (content)
            {
                return e.Values.Select(x => CreateNodes(x, false, "")[0]).ToArray();
            }
            else
            {
                TreeViewItem item = new TreeViewItem();
                item.Header = prefix + "[ArrayInitializationExpression]";
                if (e.Values.Count != 0) { SetExpandable(item, e); }
                return new TreeViewItem[] { item };
            }
        }
        private static TreeViewItem[] CreateSetInitializationExpression(SetInitializationExpression e, bool content, string prefix)
        {
            if (content)
            {
                return e.Values.Select(x => CreateNodes(x, false, "")[0]).ToArray();
            }
            else
            {
                TreeViewItem item = new TreeViewItem();
                item.Header = prefix + "[SetInitializationExpression]";
                if (e.Values.Count != 0) { SetExpandable(item, e); }
                return new TreeViewItem[] { item };
            }
        }

        private static TreeViewItem[] CreateArrayRange(ArrayRangeExpression e, bool content, string prefix)
        {
            if (content)
            {
                return new TreeViewItem[]{
                    CreateNodes(e.Array, false, "Array ")[0],
                    CreateNodes(e.Start, false, "Start ")[0],
                    CreateNodes(e.End, false, "End ")[0],
                };
            }
            else
            {
                TreeViewItem item = new TreeViewItem();
                item.Header = prefix + "[ArrayRangeExpression]";
                SetExpandable(item, e);
                return new TreeViewItem[] { item };
            }
        }
        private static TreeViewItem[] CreateArrayAccessor(ArrayExpression e, bool content, string prefix)
        {
            if (content)
            {
                return new TreeViewItem[]{
                    CreateNodes(e.Array, false, "Array ")[0],
                    CreateNodes(e.Index, false, "Index ")[0]
                };
            }
            else
            {
                TreeViewItem item = new TreeViewItem();
                item.Header = prefix + "[ArrayExpression]";
                SetExpandable(item, e);
                return new TreeViewItem[] { item };
            }
        }
        private static TreeViewItem[] CreateFunctionCall(CallExpression e, bool content, string prefix)
        {
            if (content) {
                TreeViewItem args = new TreeViewItem();
                for (int i = 0; i < e.Arguments.Count; ++i) {
                    args.Items.Add(CreateNodes(e.Arguments[i], false, "")[0]);
                }
                args.Header = "[Arguments]";
                return new TreeViewItem[]{
                    CreateNodes(e.Function, false, "Function ")[0],
                    args
                };
            }
            else
            {
                TreeViewItem item = new TreeViewItem();
                item.Header = prefix + "[CallExpression]";
                SetExpandable(item, e);
                return new TreeViewItem[] { item };
            }
        }
        private static TreeViewItem[] CreateMemberAccessor(MemberAccessor ma, bool content, string prefix)
        {
            if (content)
            {
                return new TreeViewItem[]{
                    CreateNodes(ma.Object, false, "Object ")[0],
                    CreateNodes(ma.Member, false, "Member ")[0]
                };
            }
            else
            {
                TreeViewItem item = new TreeViewItem();
                item.Header = prefix + "[MemberAccessor]";
                SetExpandable(item, ma);
                return new TreeViewItem[] { item };
            }
        }
        private static TreeViewItem[] CreateGroup(GroupExpression e, bool content, string prefix)
        {
            if (content)
            {
                return CreateNodes(e.Expression, false, "");
            }
            else
            {
                TreeViewItem item = new TreeViewItem();
                item.Header = prefix + "[GroupExpression " + e.GroupType + "]";
                SetExpandable(item, e);
                return new TreeViewItem[] { item };
            }
        }
        private static TreeViewItem[] CreateBinary(BinaryExpression e, bool content, string prefix)
        {
            if (content)
            {
                return new TreeViewItem[]{
                    CreateNodes(e.Left, false, "Left ")[0],
                    CreateNodes(e.Right, false, "Right ")[0]
                };
            }
            else
            {
                TreeViewItem item = new TreeViewItem();
                item.Header = prefix + "[BinaryExpression " + e.Operator + "]";
                SetExpandable(item, e);
                return new TreeViewItem[] { item };
            }
        }
        private static TreeViewItem[] CreateUnary(UnaryExpression e, bool content, string prefix)
        {
            if (content)
            {
                return CreateNodes(e.Argument, false, "");
            }
            else
            {
                TreeViewItem item = new TreeViewItem();
                item.Header = prefix + "[UnaryExpression " + e.Operator + "]";
                SetExpandable(item, e);
                return new TreeViewItem[] { item };
            }
        }

        private static TreeViewItem[] CreateProgram(ProgramStatement s, bool content, string prefix)
        {
            if (content)
            {
                TreeViewItem[] items = new TreeViewItem[s.Body.Count];
                for (int i = 0; i < items.Length; ++i)
                {
                    items[i] = CreateNodes(s.Body[i], false, "")[0];
                }
                return items;
            }
            else
            {
                TreeViewItem item = new TreeViewItem();
                item.Header = prefix + "[ProgramStatement]";
                SetExpandable(item, s);
                return new TreeViewItem[] { item };
            }
        }
        private static TreeViewItem[] CreateFunctionDefinition(FunctionDeclaration s, bool content, string prefix)
        {
            if (content)
            {
                TreeViewItem paramsItem = new TreeViewItem();
                paramsItem.Header = "[Parameters]";
                foreach (Identifier i in s.Parameters) {
                    paramsItem.Items.Add(CreateNodes(i, false, "")[0]);
                }
                return new TreeViewItem[]{
                    paramsItem,
                    CreateNodes(s.Body, false, "Body ")[0]
                };
            }
            else
            {
                TreeViewItem item = new TreeViewItem();
                item.Header = prefix + "[FunctionDeclaration "+s.Name.Name+"]";
                SetExpandable(item, s);
                return new TreeViewItem[] { item };
            }
        }
        private static TreeViewItem[] CreateExpression(ExpressionStatement s, bool content, string prefix)
        {
            if (content)
            {
                return new TreeViewItem[]{
                    CreateNodes(s.Expression, false, "")[0]
                };
            }
            else
            {
                TreeViewItem item = new TreeViewItem();
                item.Header = prefix + "[ExpressionStatement]";
                SetExpandable(item, s);
                return new TreeViewItem[] { item };
            }
        }
        private static TreeViewItem[] CreateBlock(BlockStatement s, bool content, string prefix)
        {
            if (content)
            {
                TreeViewItem[] items = new TreeViewItem[s.Body.Count];
                for (int i = 0; i < items.Length; ++i) {
                    items[i] = CreateNodes(s.Body[i], false, "")[0];
                }
                return items;
            }
            else
            {
                TreeViewItem item = new TreeViewItem();
                item.Header = prefix + "[BlockStatement]";
                SetExpandable(item, s);
                return new TreeViewItem[] { item };
            }
        }
        private static TreeViewItem[] CreateReturn(ReturnStatement s, bool content, string prefix)
        {
            if (content)
            {
                return new TreeViewItem[]{
                    CreateNodes(s.Argument, false, "")[0]
                };
            }
            else
            {
                TreeViewItem item = new TreeViewItem();
                item.Header = prefix + "[ReturnStatement]";
                SetExpandable(item, s);
                return new TreeViewItem[] { item };
            }
        }
        private static TreeViewItem[] CreatePrint(PrintStatement s, bool content, string prefix)
        {
            if (content)
            {
                return new TreeViewItem[]{
                    CreateNodes(s.Argument, false, "")[0]
                };
            }
            else
            {
                TreeViewItem item = new TreeViewItem();
                item.Header = prefix + "[PrintStatement]";
                SetExpandable(item, s);
                return new TreeViewItem[] { item };
            }
        }
        private static TreeViewItem[] CreateSwap(SwapStatement s, bool content, string prefix)
        {
            if (content)
            {
                return new TreeViewItem[]{
                    CreateNodes(s.Left, false, "Left ")[0],
                    CreateNodes(s.Right, false, "Right ")[0]
                };
            }
            else
            {
                TreeViewItem item = new TreeViewItem();
                item.Header = prefix + "[SwapStatement]";
                SetExpandable(item, s);
                return new TreeViewItem[] { item };
            }
        }
        private static TreeViewItem[] CreateIf(IfStatement s, bool content, string prefix)
        {
            if (content)
            {
                return new TreeViewItem[]{
                    CreateNodes(s.Test, false, "Test ")[0],
                    CreateNodes(s.Consequent, false, "Consequent ")[0],
                    CreateNodes(s.Alternate, false, "Alternate ")[0]
                };
            }
            else
            {
                TreeViewItem item = new TreeViewItem();
                item.Header = prefix + "[IfStatement]";
                SetExpandable(item, s);
                return new TreeViewItem[] { item };
            }
        }
        private static TreeViewItem[] CreateFor(ForStatement s, bool content, string prefix)
        {
            if (content)
            {
                return new TreeViewItem[]{
                    new TreeViewItem(){ Header="IterationType "+s.IterationType },
                    CreateNodes(s.Initial, false, "Initial ")[0],
                    CreateNodes(s.End, false, "End ")[0],
                    CreateNodes(s.Body, false, "Body ")[0]
                };
            }
            else
            {
                TreeViewItem item = new TreeViewItem();
                item.Header = prefix + "[ForStatement]";
                SetExpandable(item, s);
                return new TreeViewItem[] { item };
            }
        }
        private static TreeViewItem[] CreateForEach(ForEachStatement s, bool content, string prefix)
        {
            if (content)
            {
                return new TreeViewItem[]{
                    CreateNodes(s.IterationVariable, false, "IterationOver ")[0],
                    CreateNodes(s.Iterable, false, "Iterable ")[0],
                    CreateNodes(s.Body, false, "Body ")[0]
                };
            }
            else
            {
                TreeViewItem item = new TreeViewItem();
                item.Header = prefix + "[ForEachStatement]";
                SetExpandable(item, s);
                return new TreeViewItem[] { item };
            }
        }
        private static TreeViewItem[] CreateWhile(WhileStatement s, bool content, string prefix)
        {
            if (content)
            {
                return new TreeViewItem[]{
                    CreateNodes(s.Test, false, "Test ")[0],
                    CreateNodes(s.Body, false, "Body ")[0]
                };
            }
            else
            {
                TreeViewItem item = new TreeViewItem();
                item.Header = prefix + "[WhileStatement]";
                SetExpandable(item, s);
                return new TreeViewItem[] { item };
            }
        }
        private static TreeViewItem[] CreateRepeat(RepeatStatement s, bool content, string prefix)
        {
            if (content)
            {
                return new TreeViewItem[]{
                    CreateNodes(s.Body, false, "Body ")[0],
                    CreateNodes(s.Test, false, "Test ")[0]
                };
            }
            else
            {
                TreeViewItem item = new TreeViewItem();
                item.Header = prefix + "[RepeatStatement]";
                SetExpandable(item, s);
                return new TreeViewItem[] { item };
            }
        }

        private static void SetExpandable(TreeViewItem item, Node node) {
            item.Tag = node;
            item.Expanded += TreeViewItem_Expanded;
            item.Items.Add(new TreeViewItem());//dummy object
        }

        private static void TreeViewItem_Expanded(object sender, RoutedEventArgs e)
        {
            TreeViewItem item = (TreeViewItem)sender;
            item.Expanded -= TreeViewItem_Expanded;
            TreeViewItem[] items = CreateNodes((Node)item.Tag, true, "");
            item.Items.Clear();
            foreach (var i in items) {
                item.Items.Add(i);
            }
        }
    }
}
