﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using PseudoInterpreter.Recursive;
using PseudoInterpreter.Interpreter;
using System.Numerics;

namespace PseudoInterpreter.UI
{
    internal sealed class InterpreterEnvironment
    {
        private Action<string> Write;
        public InterpreterEnvironment(Action<string> write) {
            Write = write;
        }

        public Provider Time()
        {
            return new IntegerConstantProvider((BigInteger)(DateTime.UtcNow - new DateTime(1970, 1, 1)).TotalMilliseconds);
        }
        public IEnumerable<Result<Provider>> Min(AstInterpreter ai, ReferenceValue[] arguments)
        {
            if (arguments.Length == 0) { yield return Constants.Nil; }

            ReferenceValue min = arguments[0];
            Result<bool> result = new Result<bool>();

            for (int i = 1; i < arguments.Length; ++i)
            {
                yield return ai.Call<Provider, bool>(min.IsGreater(ai, arguments[i]), result);
                if ((bool)result) { min = arguments[i]; }
            }
            yield return min.Provider;
        }
        public IEnumerable<Result<Provider>> Max(AstInterpreter ai, ReferenceValue[] arguments)
        {
            if (arguments.Length == 0) { yield return Constants.Nil; }

            ReferenceValue max = arguments[0];
            Result<bool> result = new Result<bool>();

            for (int i = 1; i < arguments.Length; ++i)
            {
                yield return ai.Call<Provider, bool>(max.IsLess(ai, arguments[i]), result);
                if ((bool)result) { max = arguments[i]; }
            }
            yield return max.Provider;
        }
        public IEnumerable<Result<Provider>> Dump(AstInterpreter ai, ReferenceValue[] arguments)
        {
            if (arguments.Length == 0) { yield return Constants.Nil; }

            Result<string> result = new Result<string>();
            yield return ai.Call<Provider, string>(DumpToString(ai, arguments[0].Provider, new Dictionary<Provider, int>(), 1), result);
            Write(result + "\n");
            yield return Constants.Nil;
        }
        public IEnumerable<Result> Print(AstInterpreter ai, ReferenceValue value)
        {
            Result<string> str = new Result<string>();
            yield return ai.Call<string>(PrintToString(ai, value.Provider, new Dictionary<Provider, int>(), 1), str);
            Write(str + "\n");
        }
        private IEnumerable<Result<string>> PrintToString(AstInterpreter ai, Provider p, Dictionary<Provider, int> stackProvider, int currentLevel)
        {
            string result = "";
            if (p == Constants.NaN) { result = "NaN"; }
            else if (p == Constants.NegativeInfinity) { result = "-Infinity"; }
            else if (p == Constants.Nil) { result = "Nil"; }
            else if (p == Constants.PositiveInfinity) { result = "Infinity"; }
#if TYPE_UNDEFINED
            else if (p == Constants.Undefined) { result = "Undefinied"; }
#endif
            else
            {
                //number, set, array, object
                Type type = p.GetType();
                if (type == typeof(FloatConstantProvider))
                {
                    result = ((FloatConstantProvider)p).Value.ToString();
                }
                else if (type == typeof(IntegerConstantProvider))
                {
                    result = ((IntegerConstantProvider)p).Value.ToString();
                }
                else if (type == typeof(SetProvider))
                {
                    SetProvider sp = (SetProvider)p;
                    if (stackProvider.ContainsKey(sp))
                    {
                        result += "{# Reference " + stackProvider[sp] + " #}";
                    }
                    else
                    {
                        stackProvider.Add(sp, currentLevel);
                        result = "{ ";
                        Result<string> str = new Result<string>();
                        bool firstValue = true;
                        foreach (Provider pValue in sp.Value)
                        {
                            if (firstValue) { firstValue = false; }
                            else { result += ", "; }
                            yield return ai.Call<string, string>(PrintToString(ai, pValue, stackProvider, currentLevel + 1), str);
                            result += str;
                        }
                        result += " }";
                        stackProvider.Remove(sp);
                    }
                }
                else if (type == typeof(ArrayProvider))
                {
                    ArrayProvider ap = (ArrayProvider)p;
                    if (stackProvider.ContainsKey(ap))
                    {
                        result += "[# Reference " + stackProvider[ap] + " #]";
                    }
                    else
                    {
                        stackProvider.Add(ap, currentLevel);
                        result = "[ ";
                        Result<string> str = new Result<string>();
                        for (int i = 1; i <= ap.Length; ++i)
                        {
                            if (i > 1) { result += ", "; }
                            if (ap.Value.ContainsKey(i))
                            {
                                yield return ai.Call<string, string>(PrintToString(ai, ap.Value[i].Provider, stackProvider, currentLevel + 1), str);
                                result += str;
                            }
                            else { result += "Undefined"; }
                        }
                        result += " ]";
                        stackProvider.Remove(ap);
                    }
                }
                else if (type == typeof(ObjectProvider))
                {
                    ObjectProvider op = (ObjectProvider)p;
                    if (stackProvider.ContainsKey(op))
                    {
                        result += "{:# Reference " + stackProvider[op] + " #:}";
                    }
                    else
                    {
                        stackProvider.Add(op, currentLevel);
                        result = "{: ";
                        Result<string> str = new Result<string>();
                        bool firstValue = true;
                        foreach (KeyValuePair<string, ReferenceValue> kvp in op.Value)
                        {
                            if (firstValue) { firstValue = false; }
                            else { result += ", "; }
                            yield return ai.Call<string, string>(PrintToString(ai, kvp.Value.Provider, stackProvider, currentLevel + 1), str);
                            result += "\"" + kvp.Key + "\": " + str;
                        }
                        result += " :}";
                        stackProvider.Remove(op);
                    }
                }
                else if (type == typeof(NativeMethodProvider)) { result = "#NativeMethod#"; }
                else if (type == typeof(FunctionProvider))
                {
                    FunctionProvider fp = (FunctionProvider)p;
                    result = fp.MethodInfo.Name + "(";
                    bool firstValue = true;
                    foreach (string pname in fp.MethodInfo.Parameters)
                    {
                        if (firstValue) { firstValue = false; }
                        else { result += ", "; }
                        result += pname;
                    }
                    result += ")";
                }
            }
            yield return result;
        }
        private IEnumerable<Result<string>> DumpToString(AstInterpreter ai, Provider p, Dictionary<Provider, int> stackProvider, int currentLevel)
        {
            string result = "";
            if (p == Constants.NaN) { result = "[Special NaN]"; }
            else if (p == Constants.NegativeInfinity) { result = "[Special -Infinity]"; }
            else if (p == Constants.Nil) { result = "[Nil]"; }
            else if (p == Constants.PositiveInfinity) { result = "[Special Infinity]"; }
#if TYPE_UNDEFINED
            else if (p == Constants.Undefined) { result = "[Undefinied]"; }
#endif
            else
            {
                //number, set, array, object
                Type type = p.GetType();
                if (type == typeof(FloatConstantProvider))
                {
                    result = "[Float: " + ((FloatConstantProvider)p).Value.ToString() + "]";
                }
                else if (type == typeof(IntegerConstantProvider))
                {
                    result = "[Integer: " + ((IntegerConstantProvider)p).Value.ToString() + "]";
                }
                else if (type == typeof(SetProvider))
                {
                    SetProvider sp = (SetProvider)p;
                    if (stackProvider.ContainsKey(sp))
                    {
                        result += "[Reference: " + stackProvider[sp] + "]";
                    }
                    else
                    {
                        stackProvider.Add(sp, currentLevel);
                        result = "[Set: ";
                        Result<string> str = new Result<string>();
                        bool firstValue = true;
                        foreach (Provider pValue in sp.Value)
                        {
                            if (firstValue) { firstValue = false; }
                            else { result += ", "; }
                            yield return ai.Call<string, string>(DumpToString(ai, pValue, stackProvider, currentLevel + 1), str);
                            result += str;
                        }
                        firstValue = true;
                        foreach (KeyValuePair<string, ReferenceValue> kvp in sp.Property)
                        {
                            if (firstValue) { firstValue = false; }
                            else { result += ", "; }
                            yield return ai.Call<string, string>(DumpToString(ai, kvp.Value.Provider, stackProvider, currentLevel + 1), str);
                            result += "{" + kvp.Key + ": " + str + "}";
                        }
                        result += "]";
                        stackProvider.Remove(sp);
                    }
                }
                else if (type == typeof(ArrayProvider))
                {
                    ArrayProvider ap = (ArrayProvider)p;
                    if (stackProvider.ContainsKey(ap))
                    {
                        result += "[Reference: " + stackProvider[ap] + "]";
                    }
                    else
                    {
                        stackProvider.Add(ap, currentLevel);
                        result = "[Array(" + ap.Length + "): ";
                        Result<string> str = new Result<string>();
                        for (int i = 1; i <= ap.Length; ++i)
                        {
                            if (i > 1) { result += ", "; }
                            if (ap.Value.ContainsKey(i))
                            {
                                yield return ai.Call<string, string>(DumpToString(ai, ap.Value[i].Provider, stackProvider, currentLevel + 1), str);
                                result += "{ " + i + ": " + str + " }";
                            }
                            else { result += "{ " + i + " : [Undefined] }"; }
                        }
                        bool firstValue = true;
                        foreach (KeyValuePair<string, ReferenceValue> kvp in ap.Property)
                        {
                            if (firstValue) { firstValue = false; }
                            else { result += ", "; }
                            yield return ai.Call<string, string>(DumpToString(ai, kvp.Value.Provider, stackProvider, currentLevel + 1), str);
                            result += "{ " + kvp.Key + ": " + str + " }";
                        }
                        result += " ]";
                        stackProvider.Remove(ap);
                    }
                }
                else if (type == typeof(ObjectProvider))
                {
                    ObjectProvider op = (ObjectProvider)p;
                    if (stackProvider.ContainsKey(op))
                    {
                        result += "[Reference: " + stackProvider[op] + "]";
                    }
                    else
                    {
                        stackProvider.Add(op, currentLevel);
                        result = "[Object: ";
                        Result<string> str = new Result<string>();
                        bool firstValue = true;
                        foreach (KeyValuePair<string, ReferenceValue> kvp in op.Value)
                        {
                            if (firstValue) { firstValue = false; }
                            else { result += ", "; }
                            yield return ai.Call<string, string>(DumpToString(ai, kvp.Value.Provider, stackProvider, currentLevel + 1), str);
                            result += "{ " + kvp.Key + ": " + str + " }";
                        }
                        result += " ]";
                        stackProvider.Remove(op);
                    }
                }
                else if (type == typeof(NativeMethodProvider)) { result = "[NativeMethod]"; }
                else if (type == typeof(FunctionProvider))
                {
                    FunctionProvider fp = (FunctionProvider)p;
                    result = "[Function " + fp.MethodInfo.Name + "(";
                    bool firstValue = true;
                    foreach (string pname in fp.MethodInfo.Parameters)
                    {
                        if (firstValue) { firstValue = false; }
                        else { result += ", "; }
                        result += pname;
                    }
                    result += ") ]";
                }
            }
            yield return result;
        }

    }
}
