﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Threading.Tasks;

using PseudoInterpreter.Tokenizer;
using PseudoInterpreter.Recursive;

namespace PseudoInterpreter.Ast
{
    public sealed class AstParser : StackMachine
    {
        //Sintassi dei commenti all'interno dei parse....
        //{something} is result a parsed node of type {something}
        //{something!} is a left assignable something
        //[something] is a token of type [something]
        //( starts a group, ) ends a group
        //? optional operator
        //+ 1 or more occurence
        //* 0 or more occurence
        //{a,b} from a to b occurence (both inclusive), b can be * for no upper bound

        private Tokenizer.Tokenizer Tokenizer;
        private Token[] Tokens;
        private int Index = 0;
        private Exception ParseError;

        public AstParser(TextReader reader)
        {
            Tokenizer = new Tokenizer.Tokenizer(reader);
        }

        public ProgramStatement ParseTree()
        {
            Tokens = Tokenizer.Tokenize().ToArray();
            Index = 0;
            List<Statement> stats = new List<Statement>();
            while (Index < Tokens.Length)
            {
                if (Tokens[Index].TokenType == TokenType.EndInput) { break; }
                Statement stat = CallImmediate(ParseStatement(true, false));
                if (stat == null)
                {
                    throw ParseError ?? new InvalidOperationException("![Ops... This exceptional situation was not expected]");
                }
                stats.Add(stat);
            }
            return new ProgramStatement(stats);
        }


        private ReturnStatement ParseReturnStatement()
        {
            //[Return] {Expression}? [EndStatement]
            int oldIndex = Index;
            Expression argument;

            if (Tokens[Index].TokenType != TokenType.Return)
            {
                ErrorTokenMismatch(TokenType.Remainder, Tokens[Index], Index);
                return null;
            }
            ++Index;

            argument = CallImmediate<Expression>(ParseExpression());
            if (Tokens[Index].TokenType != TokenType.EndStatement)
            {
                ErrorTokenMismatch(TokenType.EndStatement, Tokens[Index], Index);
                Index = oldIndex;
                return null;
            }
            ++Index;
            return new ReturnStatement(argument);
        }
        private PrintStatement ParsePrintStatement()
        {
            //[Print] {Expression} [EndStatement]
            int oldIndex = Index;
            Expression argument;

            if (Tokens[Index].TokenType != TokenType.Print) { return null; }
            ++Index;

            argument = CallImmediate<Expression>(ParseExpression());
            if (argument == null)
            {
                ErrorExpressionExpected(Tokens[Index], Index);
                Index = oldIndex;
                return null;
            }
            if (Tokens[Index].TokenType != TokenType.EndStatement)
            {
                ErrorTokenMismatch(TokenType.EndStatement, Tokens[Index], Index);
                Index = oldIndex;
                return null;
            }
            ++Index;
            return new PrintStatement(argument);
        }
        private SwapStatement ParseSwapStatement()
        {
            //[Swap] {Expression!} [DoubleArrow] {Expression!} [EndStatement]
            int oldIndex = Index;
            Expression left, right;

            if (Tokens[Index].TokenType != TokenType.Swap) { return null; }
            ++Index;

            left = CallImmediate<Expression>(ParseExpression());
            if (left == null)
            {
                ErrorExpressionExpected(Tokens[Index], Index);
                Index = oldIndex;
                return null;
            }
            if (!left.IsLeftAssignable && !(left is BinaryExpression) && ((BinaryExpression)left).Operator != BinaryOperator.Assign)
            {
                Index = oldIndex;
                Error("Left operand of swap must be assignable", oldIndex);
                return null;
            }
            if (Tokens[Index].TokenType != TokenType.DoubleArrow)
            {
                Index = oldIndex;
                ErrorTokenMismatch(TokenType.DoubleArrow, Tokens[Index], Index);
                return null;
            }
            ++Index;//consume <->

            right = CallImmediate<Expression>(ParseExpression());
            if (right == null)
            {
                ErrorExpressionExpected(Tokens[Index], Index);
                Index = oldIndex;
                return null;
            }
            if (!right.IsLeftAssignable && !(right is BinaryExpression) && ((BinaryExpression)right).Operator != BinaryOperator.Assign)
            {
                Index = oldIndex;
                Error("Right operand of swap must be assignable", oldIndex);
                return null;
            }
            //we have read all info necessary

            if (Tokens[Index].TokenType != TokenType.EndStatement)
            {
                ErrorTokenMismatch(TokenType.EndStatement, Tokens[Index], Index);
                Index = oldIndex;
                return null;
            }
            ++Index;
            return new SwapStatement(left, right);
        }
        private ExpressionStatement ParseExpressionStatement()
        {
            int oldIndex = Index;
            Expression expr = CallImmediate<Expression>(ParseExpression());
            if (Tokens[Index].TokenType != TokenType.EndStatement)
            {
                ErrorTokenMismatch(TokenType.EndStatement, Tokens[Index], Index);
                Index = oldIndex;
                return null;
            }
            ++Index;
            return new ExpressionStatement(expr);
        }

        private IEnumerable<Result<WhileStatement>> ParseWhileStatement()
        {
            //[While] {Expression} [EndStatement]
            //{BlockStatement}?
            int oldIndex = Index;
            Expression test;
            Result<BlockStatement> block = new Result<BlockStatement>();

            if (Tokens[Index].TokenType != TokenType.While)
            {
                ErrorTokenMismatch(TokenType.While, Tokens[Index], Index);
                yield return null;
            }
            ++Index;
            test = CallImmediate(ParseExpression());
            if (test == null)
            {
                Index = oldIndex;
                yield return null;
            }
            if (Tokens[Index].TokenType != TokenType.EndStatement)
            {
                ErrorTokenMismatch(TokenType.EndStatement, Tokens[Index], Index);
                Index = oldIndex;
                yield return null;
            }
            ++Index;
            //parse body
            yield return Call<WhileStatement, BlockStatement>(ParseBlockStatement(), block);
            yield return new WhileStatement(test, block);
        }
        private IEnumerable<Result<RepeatStatement>> ParseRepeatStatement()
        {
            //Ha due forme

            //[Repeat] {Statement}
            //[Until] {Expression}

            //[Repeat] [EndStatement]
            //{BlockStatement}?
            //[Until] {Expression}

            int oldIndex = Index;
            Expression test;
            Statement body = null;

            if (Tokens[Index].TokenType != TokenType.Repeat)
            {
                ErrorTokenMismatch(TokenType.Repeat, Tokens[Index], Index);
                yield return null;
            }
            ++Index;
            if (Tokens[Index].TokenType != TokenType.EndStatement)
            {
                Result<Statement> result = new Result<Statement>();
                yield return Call<RepeatStatement, Statement>(ParseStatement(), result);
                body = result;
                if (body == null)
                {
                    Index = oldIndex;
                    ErrorExpressionExpected(Tokens[Index], Index);
                    yield return null;
                }
            }
            else
            {
                ++Index;
                Result<BlockStatement> result = new Result<BlockStatement>();
                yield return Call<RepeatStatement, BlockStatement>(ParseBlockStatement(), result);
                body = result;
            }
            //body can be null here
            //now parse until
            if (Tokens[Index].TokenType != TokenType.Until)
            {
                Index = oldIndex;
                ErrorTokenMismatch(TokenType.Until, Tokens[Index], Index);
                yield return null;
            }
            ++Index;//consume until
            test = CallImmediate(ParseExpression());
            if (test == null)
            {
                Index = oldIndex;
                ErrorExpressionExpected(Tokens[Index], Index);
                yield return null;
            }
            if (Tokens[Index].TokenType != TokenType.EndStatement)
            {
                Index = oldIndex;
                ErrorTokenMismatch(TokenType.EndStatement, Tokens[Index], Index);
                yield return null;
            }
            ++Index;
            //return repeat statement
            yield return new RepeatStatement(test, body);
        }
        private IEnumerable<Result<BlockStatement>> ParseBlockStatement()
        {
            //[StartBlock] {Statement}* [EndBlock]

            int oldIndex = Index;
            BlockStatement block = new BlockStatement();

            if (Tokens[Index].TokenType != TokenType.StartBlock)
            {
                ErrorTokenMismatch(TokenType.StartBlock, Tokens[Index], Index);
                yield return null;
            }
            ++Index;

            while (Tokens[Index].TokenType != TokenType.EndBlock)
            {
                Result<Statement> statement = new Result<Statement>();
                yield return Call<BlockStatement, Statement>(ParseStatement(), statement);
                if ((Statement)statement == null)
                {
                    Index = oldIndex;
                    yield return null;
                }
                block.Body.Add(statement);
            }
            ++Index;
            yield return block;
        }
        private IEnumerable<Result<Statement>> ParseForStatement()
        {
            //Ha un po' di forme
            //[For] {Expression!} [To] {Expression} [EndStatement]
            //{BlockStatement}?

            //[For] {Expression!} [Downto] {Expression} [EndStatement]
            //{BlockStatement}?

            //[For] [Each] {Expression!} [ElementOf] {Expression} [EndStatement]
            //{BlockStatement}?

            int oldIndex = Index;
            Expression start, end;
            ForIterationType iterationType = ForIterationType.Upto;
            Result<BlockStatement> block = new Result<BlockStatement>();

            //if it doesn't find for keyword it's not a for loop
            if (Tokens[Index].TokenType != TokenType.For)
            {
                ErrorTokenMismatch(TokenType.For, Tokens[Index], Index);
                yield return null;
            }
            ++Index;
            if (Tokens[Index].TokenType == TokenType.Each)
            {
                //for each statement....
                ++Index;
                start = CallImmediate(ParseExpression());
                if (start == null || !(start is BinaryExpression) || ((BinaryExpression)start).Operator != BinaryOperator.ElementOf)
                {
                    Error("ForEach initial expression must be a membership", oldIndex);
                    Index = oldIndex;
                    yield return null;
                }
                if (!((BinaryExpression)start).Left.IsLeftAssignable)
                {
                    Error("Left expression of membership must be assignable", oldIndex);
                    Index = oldIndex;
                    yield return null;
                }
                end = ((BinaryExpression)start).Right;
                start = ((BinaryExpression)start).Left;
                //for end
                if (Tokens[Index].TokenType != TokenType.EndStatement)
                {
                    ErrorTokenMismatch(TokenType.EndStatement, Tokens[Index], Index);
                    Index = oldIndex;
                    yield return null;
                }
                ++Index;
                yield return Call<Statement, BlockStatement>(ParseBlockStatement(), block);
                yield return new ForEachStatement(start, end, block);
            }
            else
            {
                start = CallImmediate(ParseExpression());
                if (start == null || (!start.IsLeftAssignable && (!(start is BinaryExpression) || ((BinaryExpression)start).Operator != BinaryOperator.Assign)))
                {
                    Error("For initial expression must be an assignment or left assignable", oldIndex);
                    Index = oldIndex;
                    yield return null;
                }

                //expression not found or {down}?to keyword not found
                if (Tokens[Index].TokenType == TokenType.To)
                {
                    iterationType = ForIterationType.Upto;
                }
                else if (Tokens[Index].TokenType == TokenType.Downto)
                {
                    iterationType = ForIterationType.Downto;
                }
                else
                {
                    ErrorTokenMismatch(TokenType.To, Tokens[Index], Index);
                    Index = oldIndex;
                    yield return null;
                }
                ++Index;
                end = CallImmediate(ParseExpression());
            }
            //for end
            if (Tokens[Index].TokenType != TokenType.EndStatement)
            {
                ErrorTokenMismatch(TokenType.EndStatement, Tokens[Index], Index);
                Index = oldIndex;
                yield return null;
            }
            ++Index;
            yield return Call<Statement, BlockStatement>(ParseBlockStatement(), block);

            yield return new ForStatement(iterationType, start, end, block);
        }
        private IEnumerable<Result<IfStatement>> ParseIfStatement()
        {
            //[If] {Expression} [EndStatement]
            //{BlockStatement}?
            //[Else] {Statement}

            //[If] {Expression} [EndStatement]
            //{BlockStatement}?
            //[Else] [EndStatement]
            //{BlockStatement}?

            int oldIndex = Index;
            Expression test;
            Result<BlockStatement> consequent = new Result<BlockStatement>();
            Result<Statement> alternate = new Result<Statement>();

            if (Tokens[Index].TokenType != TokenType.If)
            {
                ErrorTokenMismatch(TokenType.If, Tokens[Index], Index);
                yield return null;
            }
            ++Index;
            test = CallImmediate(ParseExpression());
            if (test == null)
            {
                Index = oldIndex;
                yield return null;
            }
            //Then è opzionale
            if (Tokens[Index].TokenType == TokenType.Then) { ++Index; }
            if (Tokens[Index].TokenType != TokenType.EndStatement)
            {
                ErrorTokenMismatch(TokenType.EndStatement, Tokens[Index], Index);
                Index = oldIndex;
                yield return null;
            }
            ++Index;
            yield return Call<IfStatement, BlockStatement>(ParseBlockStatement(), consequent);

            if (Tokens[Index].TokenType == TokenType.Else)
            {
                ++Index;
                if (Tokens[Index].TokenType != TokenType.EndStatement)
                {
                    yield return Call<IfStatement, Statement>(ParseStatement(), alternate);
                    if ((Statement)alternate == null)
                    {
                        Index = oldIndex;
                        yield return null;
                    }
                }
                else
                {
                    ++Index;
                    //else block
                    Result<BlockStatement> alternateBlock = new Result<BlockStatement>();
                    yield return Call<IfStatement, BlockStatement>(ParseBlockStatement(), alternateBlock);
                    alternate = (BlockStatement)alternateBlock;
                }
            }
            yield return new IfStatement(test, consequent, alternate);
        }
        private IEnumerable<Result<FunctionDeclaration>> ParseFunctionDeclaration()
        {
            //[Identifier] [StartGroup] ([Identifier] [EndArgument])* [Identifier] [EndGroup] [EndStatement]
            //{BlockStatement}

            //[Identifier] [StartGroup] [EndGroup] [EndStatement]
            //{BlockStatement}

            int oldIndex = Index;
            string name;
            List<Identifier> identifiers = new List<Identifier>();
            HashSet<string> uniqueIdentifiers = new HashSet<string>();
            string identifier;
            Result<BlockStatement> block = new Result<BlockStatement>();

            if (Tokens[Index].TokenType != TokenType.Identifier)
            {
                ErrorTokenMismatch(TokenType.Identifier, Tokens[Index], Index);
                yield return null;
            }
            name = ((IdentifierToken)Tokens[Index]).Value;
            ++Index;
            if (Tokens[Index].TokenType != TokenType.StartGroup)
            {
                ErrorTokenMismatch(TokenType.StartGroup, Tokens[Index], Index);
                Index = oldIndex;
                yield return null;
            }
            ++Index;
            //arguments
            while (true)
            {
                if (Tokens[Index].TokenType == TokenType.EndGroup)
                {
                    ++Index;
                    break;
                }

                if (Tokens[Index].TokenType != TokenType.Identifier)
                {
                    ErrorTokenMismatch(TokenType.Identifier, Tokens[Index], Index);
                    Index = oldIndex;
                    yield return null;
                }

                identifier = ((IdentifierToken)Tokens[Index]).Value;
                if (uniqueIdentifiers.Contains(identifier))
                {
                    Error("Identifiers list must be unique on a function declaration", Index);
                    Index = oldIndex;
                    yield return null;
                }
                uniqueIdentifiers.Add(identifier);
                identifiers.Add(new Identifier(identifier));
                ++Index;

                if (Tokens[Index].TokenType == TokenType.EndArgument)
                {
                    ++Index;
                    if (Tokens[Index].TokenType == TokenType.EndGroup)
                    {
                        ErrorTokenMismatch(TokenType.Identifier, Tokens[Index], Index);
                        Index = oldIndex;
                        yield return null;
                    }
                }
                else if (Tokens[Index].TokenType != TokenType.EndGroup)
                {
                    ErrorTokenMismatch(TokenType.EndGroup, Tokens[Index], Index);
                    Index = oldIndex;
                    yield return null;
                }
            }
            if (Tokens[Index].TokenType != TokenType.EndStatement)
            {
                ErrorTokenMismatch(TokenType.EndStatement, Tokens[Index], Index);
                Index = oldIndex;
                yield return null;
            }
            ++Index;

            //read all arguments next pass is to read block definition (MUST BE PRESENT)
            yield return Call<FunctionDeclaration, BlockStatement>(ParseBlockStatement(), block);

            if ((BlockStatement)block == null)
            {
                Index = oldIndex;
                yield return null;
            }
            yield return new FunctionDeclaration(new Identifier(name), identifiers, block);
        }

        private IEnumerable<Result<Statement>> ParseStatement()
        {
            return ParseStatement(false, true);
        }
        private IEnumerable<Result<Statement>> ParseStatement(bool functionAllowed, bool returnAllowed)
        {
            switch (Tokens[Index].TokenType)
            {
                case TokenType.For:
                    yield return Return(ParseForStatement());
                    break;
                case TokenType.While:
                    yield return Return<Statement, WhileStatement>(ParseWhileStatement());
                    break;
                case TokenType.Repeat:
                    yield return Return<Statement, RepeatStatement>(ParseRepeatStatement());
                    break;
                case TokenType.Return:
                    if (!returnAllowed)
                    {
                        Error("Return statement is not allowed", Index);
                        yield return null;
                    }
                    yield return ParseReturnStatement();
                    break;
                case TokenType.Print:
                    yield return ParsePrintStatement();
                    break;
                case TokenType.Swap:
                    yield return ParseSwapStatement();
                    break;
                case TokenType.If:
                    yield return Return<Statement, IfStatement>(ParseIfStatement());
                    break;
                default:
                    if (functionAllowed)
                    {
                        Result<FunctionDeclaration> function = new Result<FunctionDeclaration>();
                        yield return Call<Statement, FunctionDeclaration>(ParseFunctionDeclaration(), function);
                        if ((FunctionDeclaration)function != null) { yield return (Statement)function; }
                    }
                    yield return ParseExpressionStatement();
                    break;
            }
        }

        private Constant ParseConstant()
        {
            //One of [ConstantDouble], [ConstantInteger], [Nil], [EmptySet], [Infinity]
            Token token = Tokens[Index];

            switch (token.TokenType)
            {
                case TokenType.ConstantDouble:
                    ++Index;
                    return new ConstantFloat(((FloatToken)token).Value);
                case TokenType.ConstantInteger:
                    ++Index;
                    return new ConstantInteger(((IntegerToken)token).Value);
                case TokenType.Infinity:
                    ++Index;
                    return new ConstantInfinity();
                case TokenType.Nil:
                    ++Index;
                    return new ConstantNil();
                case TokenType.EmptySet:
                    ++Index;
                    return new ConstantEmptySet();
            }

            ErrorTokenMismatch(new TokenType[] { TokenType.ConstantDouble, TokenType.ConstantInteger, TokenType.Infinity, TokenType.Nil }, Tokens[Index], Index);
            return null;
        }

        private IEnumerable<Result<GroupExpression>> ParseGroupExpression()
        {
            //Una tra
            //[StartGroup] {Expression} [EndGroup]
            //[StartFloor] {Expression} [EndFloor]
            //[StartCeiling] {Expression} [EndCeiling]

            int oldIndex = Index;
            GroupType groupType;
            TokenType expectEnd;

            switch (Tokens[Index].TokenType)
            {
                case TokenType.StartGroup:
                    groupType = GroupType.Group;
                    expectEnd = TokenType.EndGroup;
                    break;
                case TokenType.StartFloor:
                    groupType = GroupType.Floor;
                    expectEnd = TokenType.EndFloor;
                    break;
                case TokenType.StartCeiling:
                    groupType = GroupType.Ceiling;
                    expectEnd = TokenType.EndCeiling;
                    break;
                default:
                    ErrorTokenMismatch(new TokenType[]{
                        TokenType.StartGroup,
                        TokenType.StartCeiling,
                        TokenType.StartFloor
                    }, Tokens[Index], Index);
                    yield return null;
                    yield break;//syntax only...
            }
            ++Index;
            Result<Expression> expression = new Result<Expression>();
            yield return Call<GroupExpression, Expression>(ParseExpression(), expression);
            if ((Expression)expression == null)
            {
                Error("Empty group expression found", Index);
                Index = oldIndex;
                yield return null;
            }
            if (Tokens[Index].TokenType != expectEnd)
            {
                ErrorTokenMismatch(expectEnd, Tokens[Index], Index);
                Index = oldIndex;
                yield return null;
            }
            ++Index;
            yield return new GroupExpression(groupType, expression);
        }
        private IEnumerable<Result<ArrayInitializationExpression>> ParseArrayInitializationExpression()
        {
            //Una tra
            //[StartIndexer] ({Expression} [EndArgument])* {Expression} [EndIndexer]
            //[StartIndexer] [EndIndexer]
            int oldIndex = Index;
            if (Tokens[Index].TokenType != TokenType.StartIndexer)
            {
                ErrorTokenMismatch(TokenType.StartIndexer, Tokens[Index], Index);
                yield return null;
            }
            ++Index;
            ArrayInitializationExpression result = new ArrayInitializationExpression();
            Result<Expression> expr = new Result<Expression>();

            while (true)
            {
                if (Tokens[Index].TokenType == TokenType.EndIndexer)
                {
                    ++Index;
                    yield return result;
                }
                yield return Call<ArrayInitializationExpression, Expression>(ParseExpression(), expr);
                //deve sempre esserci un argomento in questo punto...
                if ((Expression)expr == null)
                {
                    Index = oldIndex;
                    ErrorExpressionExpected(Tokens[Index], Index);
                    yield return null;
                }
                result.Values.Add(expr);

                if (Tokens[Index].TokenType == TokenType.EndArgument)
                {
                    ++Index;
                    if (Tokens[Index].TokenType == TokenType.EndIndexer)
                    {
                        //errore ,] missing expression
                        Index = oldIndex;
                        ErrorExpressionExpected(Tokens[Index], Index);
                        yield return null;
                    }
                }
            }
        }
        private IEnumerable<Result<SetInitializationExpression>> ParseSetInitializationExpression()
        {
            //Una tra
            //[StartSet] ({Expression} [EndArgument])* {Expression} [EndSet]
            //[StartSet] [EndSet]
            int oldIndex = Index;
            if (Tokens[Index].TokenType != TokenType.StartSet)
            {
                ErrorTokenMismatch(TokenType.StartSet, Tokens[Index], Index);
                yield return null;
            }
            ++Index;
            SetInitializationExpression result = new SetInitializationExpression();
            Result<Expression> expr = new Result<Expression>();

            while (true)
            {
                if (Tokens[Index].TokenType == TokenType.EndSet)
                {
                    ++Index;
                    yield return result;
                }
                yield return Call<SetInitializationExpression, Expression>(ParseExpression(), expr);
                //deve sempre esserci un argomento in questo punto...
                if ((Expression)expr == null)
                {
                    Index = oldIndex;
                    ErrorExpressionExpected(Tokens[Index], Index);
                    yield return null;
                }
                result.Values.Add(expr);

                if (Tokens[Index].TokenType == TokenType.EndArgument)
                {
                    ++Index;
                    if (Tokens[Index].TokenType == TokenType.EndSet)
                    {
                        //errore ,} missing expression
                        Index = oldIndex;
                        ErrorExpressionExpected(Tokens[Index], Index);
                        yield return null;
                    }
                }
            }
        }
        private IEnumerable<Result<Expression>> ParseArrayExpression(Expression array)
        {
            //Una tra
            //[StartIndexer] {Expression} [EndIndexer]
            //[StartIndexer] {Expression} [Range] {Expression} [EndIndexer]
            int oldIndex = Index;
            Result<Expression> start = new Result<Expression>();
            Result<Expression> end = new Result<Expression>();
            Expression result = null;

            if (Tokens[Index].TokenType != TokenType.StartIndexer)
            {
                ErrorTokenMismatch(TokenType.StartIndexer, Tokens[Index], Index);
                yield return null;
            }
            ++Index;
            yield return Call<Expression, Expression>(ParseExpression(), start);

            if ((Expression)start == null)
            {
                Index = oldIndex;
                yield return null;
            }
            if (Tokens[Index].TokenType == TokenType.Range)
            {
                //range
                ++Index;
                yield return Call<Expression, Expression>(ParseExpression(), end);
                if ((Expression)end == null)
                {
                    Index = oldIndex;
                    yield return null;
                }
                result = new ArrayRangeExpression()
                {
                    Start = start,
                    End = end,
                    Array = array
                };
            }
            if (Tokens[Index].TokenType == TokenType.EndIndexer)
            {
                ++Index;
                if (result != null) { yield return result; }
                yield return new ArrayExpression()
                {
                    Array = array,
                    Index = start
                };
                yield break;//useless
            }

            ErrorTokenMismatch(new TokenType[] { TokenType.EndIndexer, TokenType.Range }, Tokens[Index], Index);
            Index = oldIndex;
            yield return null;
        }
        private IEnumerable<Result<Expression>> ParseVariableExpression()
        {
            //Questa è più complicata
            //VariableExpression indica il risultato di questa procedura
            //[Identifier] [MemberAccessor] {VariableExpression}
            //[Identifier] {ArrayExpression}
            //[Identifier] {ArrayRangeExpression}
            //[Identifier] [StartGroup] ({Expression} [EndArgument])* {Expression} [EndGroup]
            //[Identifier] [StartGroup] [EndGroup]

            int oldIndex = Index;
            bool isLeftAssignable = true;
            bool expectIdentifier = true;
            bool expectMemberAccessor = false;
            bool expectArrayIndexer = false;
            bool expectFunctionCall = false;
            Expression current = null;


            while (true)
            {
                Token token = Tokens[Index];
                if (expectIdentifier && token.TokenType == TokenType.Identifier)
                {
                    Identifier identifier = new Identifier(((IdentifierToken)token).Value);
                    ++Index;

                    if (current == null) { current = identifier; }
                    else if (current is MemberAccessor) { ((MemberAccessor)current).Member = identifier; }

                    expectIdentifier = false;
                    expectArrayIndexer = true;
                    expectMemberAccessor = true;
                    expectFunctionCall = true;
                }
                else if (expectMemberAccessor && token.TokenType == TokenType.MemberAccessor)
                {
                    ++Index;

                    current = new MemberAccessor()
                    {
                        Object = current
                    };

                    expectIdentifier = true;
                    expectMemberAccessor = false;
                    expectArrayIndexer = false;
                    expectFunctionCall = false;
                }
                else if (expectArrayIndexer && token.TokenType == TokenType.StartIndexer)
                {
                    Result<Expression> arrayResult = new Result<Expression>();
                    yield return Call<Expression, Expression>(ParseArrayExpression(current), arrayResult);
                    if ((Expression)arrayResult == null)
                    {
                        Index = oldIndex;
                        yield return null;
                    }
                    //if (array is ArrayRangeExpression) { head.IsLeftAssignable = false; }
                    current = arrayResult;

                    expectMemberAccessor = true;
                    expectIdentifier = false;
                    expectFunctionCall = true;
                }
                else if (expectFunctionCall && token.TokenType == TokenType.StartGroup)
                {
                    isLeftAssignable = false;
                    CallExpression call = new CallExpression();
                    ++Index;

                    while (true)
                    {
                        token = Tokens[Index];
                        if (token.TokenType == TokenType.EndGroup)
                        {
                            ++Index;
                            call.Function = current;
                            current = call;

                            expectIdentifier = false;
                            expectMemberAccessor = true;
                            expectArrayIndexer = true;
                            break;
                        }
                        else
                        {
                            Result<Expression> expr = new Result<Expression>();
                            yield return Call<Expression, Expression>(ParseExpression(), expr);
                            if ((Expression)expr == null)
                            {
                                Index = oldIndex;
                                yield return null;
                            }
                            token = Tokens[Index];
                            if (token.TokenType == TokenType.EndArgument)
                            {
                                ++Index;
                                if (Tokens[Index].TokenType == TokenType.EndGroup)
                                {
                                    ErrorExpressionExpected(Tokens[Index], Index);
                                    Index = oldIndex;
                                    yield return null;
                                }
                            }
                            else if (token.TokenType != TokenType.EndGroup)
                            {
                                ErrorTokenMismatch(TokenType.EndGroup, Tokens[Index], Index);
                                Index = oldIndex;
                                yield return null;
                            }

                            call.Arguments.Add(expr);
                        }
                    }
                }
                //end parse
                else
                {
                    if (current == null)
                    {
                        List<TokenType> et = new List<TokenType>();
                        if (expectArrayIndexer) { et.Add(TokenType.StartIndexer); }
                        if (expectFunctionCall) { et.Add(TokenType.StartGroup); }
                        if (expectIdentifier) { et.Add(TokenType.Identifier); }
                        if (expectMemberAccessor) { et.Add(TokenType.MemberAccessor); }
                        ErrorTokenMismatch(et.ToArray(), Tokens[Index], Index);
                        yield return null;
                    }
                    else
                    {
                        current.IsLeftAssignable = isLeftAssignable;
                        yield return current;
                        yield break;
                    }
                }
            }

        }
        private IEnumerable<Result<Expression>> ParseSimpleExpression()
        {
            int oldIndex = Index;
            Result<Expression> expression = ParseConstant();
            if ((Expression)expression != null) { yield return expression; }

            Result<ArrayInitializationExpression> arrInit = new Result<ArrayInitializationExpression>();
            yield return Call<Expression, ArrayInitializationExpression>(ParseArrayInitializationExpression(), arrInit);
            if ((ArrayInitializationExpression)arrInit != null) { yield return (Expression)arrInit; }

            Result<SetInitializationExpression> setInit = new Result<SetInitializationExpression>();
            yield return Call<Expression, SetInitializationExpression>(ParseSetInitializationExpression(), setInit);
            if ((SetInitializationExpression)setInit != null) { yield return (Expression)setInit; }

            Result<GroupExpression> groupExpr = new Result<GroupExpression>();
            yield return Call<Expression, GroupExpression>(ParseGroupExpression(), groupExpr);
            if ((GroupExpression)groupExpr != null) { yield return (Expression)groupExpr; }

            yield return Call<Expression, Expression>(ParseVariableExpression(), expression);
            if ((Expression)expression != null) { yield return expression; }
            yield return null;
        }
        private IEnumerable<Result<Expression>> ParseExpression()
        {
            int oldIndex = Index;
            List<Expression> expressionStack = new List<Expression>();
            List<BinaryOperator> operatorStack = new List<BinaryOperator>();
            bool expectSimple = true;
            bool expectOperator = false;
            while (true)
            {
                Token token = Tokens[Index];
                if (expectSimple)
                {
                    #region simpleExpression
                    Result<Expression> expression = new Result<Expression>();
                    if (token.TokenType == TokenType.LogicalNot || token.TokenType == TokenType.Minus)
                    {
                        UnaryOperator unaryOperator = UnaryOperator.LogicalNot;
                        //unary expression
                        switch (token.TokenType)
                        {
                            case TokenType.Minus:
                                unaryOperator = UnaryOperator.Opposite;
                                break;
                            case TokenType.LogicalNot:
                                unaryOperator = UnaryOperator.LogicalNot;
                                break;
                        }
                        ++Index;

                        yield return Call<Expression, Expression>(ParseSimpleExpression(), expression);
                        if ((Expression)expression == null)
                        {
                            Index = oldIndex;
                            yield return null;
                        }
                        expression = new UnaryExpression()
                        {
                            Operator = unaryOperator,
                            Argument = expression
                        };
                    }
                    else
                    {
                        yield return Call<Expression, Expression>(ParseSimpleExpression(), expression);
                        if ((Expression)expression == null)
                        {
                            Index = oldIndex;
                            yield return null;
                        }
                    }
                    expressionStack.Add(expression);
                    expectSimple = false;
                    expectOperator = true;
                    #endregion
                }
                else if (expectOperator)
                {
                    #region operator
                    BinaryOperator? op = AstParserHelper.GetBinaryOperator(token.TokenType);
                    int priority = op.HasValue ? AstParserHelper.GetOperatorPriority(op.Value) : -1;
                    BinaryExpression binaryExpr;

                    if (op.HasValue && op.Value == BinaryOperator.Assign)
                    {
                        if (!expressionStack[expressionStack.Count - 1].IsLeftAssignable)
                        {
                            Error("Left expression of assignment must be assignable", Index);
                            Index = oldIndex;
                            yield return null;
                        }
                    }
                    while (operatorStack.Count != 0)
                    {
                        binaryExpr = new BinaryExpression();
                        BinaryOperator previousOp = operatorStack[operatorStack.Count - 1];
                        int previousPriority = AstParserHelper.GetOperatorPriority(previousOp);
                        if ((previousPriority >= priority && op.HasValue && op.Value != BinaryOperator.Assign) || !op.HasValue)
                        {
                            //can simplify
                            binaryExpr.Operator = previousOp;
                            binaryExpr.Right = expressionStack[expressionStack.Count - 1];
                            binaryExpr.Left = expressionStack[expressionStack.Count - 2];
                            expressionStack.RemoveAt(expressionStack.Count - 1);
                            expressionStack[expressionStack.Count - 1] = binaryExpr;
                            operatorStack.RemoveAt(operatorStack.Count - 1);
                        }
                        else { break; }
                    }
                    if (op.HasValue)
                    {
                        operatorStack.Add(op.Value);
                        expectOperator = false;
                        expectSimple = true;
                        ++Index;
                    }
                    else
                    {
                        if (expressionStack.Count == 1)
                        {
                            //stop
                            yield return expressionStack[0];
                            yield break;
                        }
                        ErrorTokenMismatch(new TokenType[]{
                                TokenType.Assignment,
                                TokenType.Different,
                                TokenType.Div,
                                TokenType.Greater,
                                TokenType.GreaterEqual,
                                TokenType.Less,
                                TokenType.LessEqual,
                                TokenType.LogicalAnd,
                                TokenType.LogicalOr,
                                TokenType.Minus,
                                TokenType.Mult,
                                TokenType.Plus,
                                TokenType.Remainder,
                                TokenType.Union,
                                TokenType.Intersection,
                                TokenType.ElementOf,
                                TokenType.NotElementOf,
                                TokenType.SetMinus,
                                TokenType.SubsetOf,
                                TokenType.SubsetOfEqual,
                                TokenType.SupersetOf,
                                TokenType.SupersetOfEqual
                            }, Tokens[Index], Index);
                        Index = oldIndex;
                        yield return null;
                        yield break;
                    }
                    #endregion
                }
            }
        }

        #region ErrorFunctions
        private void ErrorTokenMismatch(TokenType[] expected, Token found, int index)
        {
            string result = "Expecting one of [";
            for (int i = 0; i < expected.Length - 1; ++i) { result += "[" + expected[i] + "] "; }
            result += "[" + expected[expected.Length - 1] + "]]";
            ParseError = new InvalidDataException(result + ", Found [" + found.TokenType + "], At Index " + index);
        }
        private void ErrorTokenMismatch(TokenType expected, Token found, int index)
        {
            ParseError = new InvalidDataException("Expecting [" + expected + "], Found [" + found.TokenType + "], At Index " + index);
        }
        private void Error(string str, int index)
        {
            ParseError = new InvalidDataException("Error[ " + str + " ] At Index " + index);
        }
        private void ErrorExpressionExpected(Token found, int index)
        {
            ParseError = new InvalidDataException("Expecting {Expression}, Found [" + found.TokenType + "] At Index " + index);
        }
        #endregion
    }
}
