﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Numerics;
using PseudoInterpreter.Recursive;

namespace PseudoInterpreter.Interpreter
{
    public sealed class ArrayProvider : Provider, IArrayProvider, ISetProvider
    {
        public readonly SparseArray Value;
        public readonly Dictionary<string, ReferenceValue> Property;
        private ReferenceValue LengthReference;

        /*public int Length {
            get { return length; }
            set {
                if (value < 0) { throw new InvalidOperationException("length cannot be negative"); }

                if (value == length) { return; }
                for (; length > value; --length)
                {
                    this.Value.Remove(length);
                }
                length = value;
                LengthReference.SetValue(new IntegerConstantProvider(length), false);
            }
        }*/
        public int Length {
            get { return Value.Length; }
            set { Value.Length = value; }
        }

        public ArrayProvider() : this(new Dictionary<int, ReferenceValue>(), 0) { }
        private ArrayProvider(Dictionary<int, ReferenceValue> value, int length) {
            LengthReference = new IntegerConstantProvider(length);
            LengthReference.ValueChanged += LengthReference_ValueChanged;

            this.Value = new SparseArray(value, LengthReference, length);
            this.Property = new Dictionary<string, ReferenceValue>();
            this.Property.Add("length", this.LengthReference);
        }
        #region Provider
        public override IEnumerable<Result<Provider>> Negate(AstInterpreter ai, ReferenceValue thisReference)
        {
            yield return this.Length == 0 ? Constants.One : Constants.Zero;
        }
        public override bool IsTrue(ReferenceValue thisReference)
        {
            return this.Length > 0;
        }
        #endregion

        #region IArrayProvider
        public ReferenceValue GetIndexedProperty(ReferenceValue thisReference, ReferenceValue rv)
        {
            Provider p = rv.Provider;
            Type type = p.GetType();
            if (type == typeof(IntegerConstantProvider))
            {
                int index = (int)((IntegerConstantProvider)p).Value;
                if (index <= 0) { throw new InvalidOperationException("Array indexes begins at 1"); }
                if (!this.Value.ContainsKey(index))
                {
                    if (index > this.Length)
                    {
                        this.Length = index;
                    }
#if TYPE_UNDEFINED
                    this.Value[index] = Constants.Undefined;
#else
                    this.Value[index] = Constants.Nil;
#endif
                }
                return this.Value[index];
            }
            throw new InvalidOperationException("An array index must be an integer value");
        }
        public ReferenceValue GetProperty(ReferenceValue thisReference, string name)
        {
            if (!Property.ContainsKey(name))
            {
#if TYPE_UNDEFINED
                Property[name] = Constants.Undefined;
#else
                Property[name] = Constants.Nil;
#endif
            }
            return Property[name];
        }

        public IEnumerable<Result<Provider>> Slice(AstInterpreter ai, ReferenceValue thisReference, ReferenceValue start, ReferenceValue end)
        {
            int startIndex, endIndex;
            Provider p1 = start.Provider,
                p2 = end.Provider;
            Type t1 = p1.GetType(),
                t2 = p2.GetType();
            if (t1 == typeof(IntegerConstantProvider))
            {
                startIndex = (int)((IntegerConstantProvider)p1).Value;
            }
            else
            {
                throw new InvalidOperationException("Cannot determine start index of slice");
            }
            if (t2 == typeof(IntegerConstantProvider))
            {
                endIndex = (int)((IntegerConstantProvider)p2).Value;
            }
            else if (p2 == Constants.PositiveInfinity) { endIndex = this.Length; }
            else
            {
                throw new InvalidOperationException("Cannot determine end index of slice");
            }


            if (startIndex <= 0) { throw new InvalidOperationException("Array indexes begins at 1"); }
            if (endIndex < startIndex) { throw new InvalidOperationException("Array end index is less than start"); }
            if (endIndex > this.Length) { endIndex = this.Length; }

            Dictionary<int, ReferenceValue> sliceValues = new Dictionary<int, ReferenceValue>();
            int newIndex = 0;
            for (; startIndex <= endIndex; ++startIndex, ++newIndex)
            {
                if (this.Value.ContainsKey(startIndex))
                {
                    ReferenceValue value = this.Value[startIndex];
#if TYPE_UNDEFINED
                    if (value.Provider != Constants.Undefined)
#else
                    if (value.Provider != Constants.Nil)
#endif
                    {
                        sliceValues.Add(newIndex + 1, value.Provider);
                    }
                }
            }

            yield return new ArrayProvider(sliceValues, newIndex);
        }
        #endregion

        #region Operators
        public static implicit operator ArrayProvider(List<ReferenceValue> arrayProt)
        {
            ArrayProvider provider = new ArrayProvider();
            int index = 0;
            foreach (var value in arrayProt)
            {
                ++index;
                provider.Value.Add(index, value);
            }
            provider.LengthReference.SetValue((IntegerConstantProvider)index, false);
            return provider;
        }
        public static implicit operator ArrayProvider(ReferenceValue[] arrayProt)
        {
            ArrayProvider provider = new ArrayProvider();
            int index = 0;
            foreach (var value in arrayProt)
            {
                ++index;
                provider.Value.Add(index, value);
            }
            provider.LengthReference.SetValue((IntegerConstantProvider)index, false);
            return provider;
        }
        public static implicit operator ArrayProvider(List<KeyValuePair<int, ReferenceValue>> arrayProt)
        {
            ArrayProvider provider = new ArrayProvider();
            int maxIndex = 0;
            foreach (var value in arrayProt)
            {
                if (value.Key > maxIndex) { maxIndex = value.Key; }
                provider.Value.Add(value.Key, value.Value);
            }
            provider.LengthReference.SetValue((IntegerConstantProvider)maxIndex, false);
            return provider;
        }
        public static implicit operator ArrayProvider(KeyValuePair<int, ReferenceValue>[] arrayProt)
        {
            ArrayProvider provider = new ArrayProvider();
            int maxIndex = 0;
            foreach (var value in arrayProt)
            {
                if (value.Key > maxIndex) { maxIndex = value.Key; }
                provider.Value.Add(value.Key, value.Value);
            }
            provider.LengthReference.SetValue((IntegerConstantProvider)maxIndex, false);
            return provider;
        }
        #endregion

        #region ISetProvider
        public IEnumerable<Result<bool>> Contains(AstInterpreter ai, ReferenceValue thisReference, ReferenceValue rv)
        {
            Result<bool> result = new Result<bool>();
            foreach (KeyValuePair<int, ReferenceValue> kvp in Value) {
                yield return ai.Call<bool, bool>(kvp.Value.IsEqual(ai, rv), result);
                if ((bool)result) { yield return true; }
            }
            yield return false;
        }
        //Wrapper around SetProvider
        public IEnumerable<Result<Provider>> SetDifference(AstInterpreter ai, ReferenceValue thisReference, ReferenceValue rv)
        {
            yield return ai.Return(new SetProvider(this.Value.Select(x => x.Value.Provider)).SetDifference(ai, thisReference, rv));
        }
        public IEnumerable<Result<Provider>> Intersect(AstInterpreter ai, ReferenceValue thisReference, ReferenceValue rv)
        {
            yield return ai.Return(new SetProvider(this.Value.Select(x => x.Value.Provider)).Intersect(ai, thisReference, rv));
        }
        public IEnumerable<Result<Provider>> Union(AstInterpreter ai, ReferenceValue thisReference, ReferenceValue rv)
        {
            yield return ai.Return(new SetProvider(this.Value.Select(x => x.Value.Provider)).Union(ai, thisReference, rv));
        }
        public IEnumerable<Result<int>> SetCompare(AstInterpreter ai, ReferenceValue thisReference, ReferenceValue rv)
        {
            yield return ai.Return(new SetProvider(this.Value.Select(x => x.Value.Provider)).Compare(ai, thisReference, rv));
        }
        public IEnumerable<Result<IEnumerator<Provider>>> GetEnumerator(AstInterpreter ai, ReferenceValue thisReference)
        {
            yield return new Result<IEnumerator<Provider>>(this.Value.Select(x => x.Value.Provider).Distinct().GetEnumerator());
        }
        #endregion

        private void LengthReference_ValueChanged(object sender, ValueChangedArgs e)
        {
            Provider currentValue = e.After;
            Type type = currentValue.GetType();
            if(type == typeof(IntegerConstantProvider)){
                this.Length = (int)((IntegerConstantProvider)currentValue).Value;
                return;
            }
            if (type == typeof(FloatConstantProvider)) {
                double d = ((FloatConstantProvider)currentValue).Value;
                if (Math.Truncate(d) == d) {
                    this.Length = (int)d;
                    return;
                }
            }
            //Error
            throw new InvalidOperationException("length must be an integer number");
        }
    }
}
