﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Numerics;
using PseudoInterpreter.Ast;
using PseudoInterpreter.Recursive;

namespace PseudoInterpreter.Interpreter
{
    public sealed class AstInterpreter : StackMachine
    {
        private Dictionary<string, ReferenceValue> GlobalScope;
        private Stack<RuntimeMethodInfo> FunctionScope;
        private ProgramStatement Program;
        private RuntimeMethodInfo CurrentScope;
        private Func<AstInterpreter, ReferenceValue, IEnumerable<Result>> PrintDelegate;

        internal ReferenceValue GetVariable(string name)
        {
            //since a variable is declared using accessor pattern or assignment and there is not way to
            //detect a global variable inside a function (eg global keyword) global variables
            //have precedence over locals (a bit unusual)
            if (GlobalScope.ContainsKey(name)) { return GlobalScope[name]; }
            if (CurrentScope != null && CurrentScope.Locals.ContainsKey(name))
            {
                return CurrentScope.Locals[name];
            }
            throw new Exception("Variable not found (BUG in interpreter)");
        }

        public void Expone(string globalName, Provider provider) {
            GlobalScope[globalName]= provider;
        }
        public void SetPrintBinding(Func<AstInterpreter, ReferenceValue, IEnumerable<Result>> printDelegate)
        {
            PrintDelegate = printDelegate;
        }

        internal IEnumerable<Result<ReferenceValue>> CallFunction(MethodInfo functionInfo, ReferenceValue[] arguments) {
            if (CurrentScope != null) { FunctionScope.Push(CurrentScope); }
            //Fix bug Call preserve reference
            for (int i = 0; i < arguments.Length; ++i) {
                arguments[i] = arguments[i].Clone();
            }
            //end fix
            CurrentScope = new RuntimeMethodInfo(functionInfo, arguments);
            yield return Call<ReferenceValue, bool>(InterpretStatement(functionInfo.Declaration.Body));

            ReferenceValue val = CurrentScope.Result ?? Constants.Nil;
            if (FunctionScope.Count == 0) { CurrentScope = null; }
            else { CurrentScope = FunctionScope.Pop(); }
            yield return val;
        }

        private IEnumerable<Result<bool>> InterpretStatement(Statement statement)
        {
            if (statement == null) {
                yield return false;
                yield break;
            }
            switch (statement.NodeType)
            {
                case NodeType.ForStatement:
                    yield return Return(InterpretForStatement((ForStatement)statement));
                    break;
                case NodeType.ForEachStatement:
                    yield return Return(InterpretForEachStatement((ForEachStatement)statement));
                    break;
                case NodeType.IfStatement:
                    yield return Return(InterpretIfStatement((IfStatement)statement));
                    break;
                case NodeType.WhileStatement:
                    yield return Return(InterpretWhileStatement((WhileStatement)statement));
                    break;
                case NodeType.ReturnStatement:
                    yield return Return(InterpretReturnStatement((ReturnStatement)statement));
                    break;
                case NodeType.ExpressionStatement:
                    yield return Return(InterpretExpressionStatement((ExpressionStatement)statement));
                    break;
                case NodeType.BlockStatement:
                    yield return Return(InterpretBlockStatement((BlockStatement)statement));
                    break;
                case NodeType.PrintStatement:
                    yield return Return(InterpretPrintStatement((PrintStatement)statement));
                    break;
                case NodeType.RepeatStatement:
                    yield return Return(InterpretRepeatStatement((RepeatStatement)statement));
                    break;
                case NodeType.SwapStatement:
                    yield return Return(InterpretSwapStatement((SwapStatement)statement));
                    break;
                default:
                    yield return false;
                    break;
            }
        }
        private IEnumerable<Result<bool>> InterpretForStatement(ForStatement statement)
        {
            Result<ReferenceValue> iteratorBlock = new Result<ReferenceValue>();
            ReferenceValue iterator;
            Result<ReferenceValue> end = new Result<ReferenceValue>();
            Result<bool> returnStatus = new Result<bool>();

            yield return Call<bool, ReferenceValue>(InterpretExpression(statement.Initial), iteratorBlock);
            iterator = iteratorBlock;
            Func<AstInterpreter, ReferenceValue, IEnumerable<Result<bool>>> comparator = null;
            Func<AstInterpreter, IEnumerable<Result<ReferenceValue>>> moveNext = null;

            switch (statement.IterationType)
            {
                case ForIterationType.Downto:
                    comparator = iterator.IsGreaterEqual;
                    moveNext = iterator.Decrement;
                    break;
                case ForIterationType.Upto:
                    comparator = iterator.IsLessEqual;
                    moveNext = iterator.Increment;
                    break;
            }
            while (true)
            {
                yield return Call<bool, ReferenceValue>(InterpretExpression(statement.End), end);
                yield return Call<bool, bool>(comparator(this, end), returnStatus);
                if ((bool)returnStatus)
                {
                    yield return Call<bool, bool>(InterpretStatement(statement.Body), returnStatus);
                    if ((bool)returnStatus)
                    {
                        yield return true;
                        yield break;
                    }
                    yield return Call<bool, ReferenceValue>(moveNext(this), iteratorBlock);
                    iterator.SetValue(iteratorBlock);
                }
                else
                {
                    yield return false;
                    yield break;
                }
            }
        }
        private IEnumerable<Result<bool>> InterpretForEachStatement(ForEachStatement statement)
        {
            Result<ReferenceValue> iteratorBlock = new Result<ReferenceValue>();
            ReferenceValue iterator;
            Result<IEnumerator<Provider>> iterableResult = new Result<IEnumerator<Provider>>();
            IEnumerator<Provider> iterable;
            Result<bool> returnStatus = new Result<bool>();

            yield return Call<bool, ReferenceValue>(InterpretExpression(statement.IterationVariable), iteratorBlock);
            iterator = iteratorBlock;
            
            yield return Call<bool, ReferenceValue>(InterpretExpression(statement.Iterable), iteratorBlock);
            yield return Call<bool, IEnumerator<Provider>>(((ReferenceValue)iteratorBlock).GetEnumerator(this), iterableResult);

            iterable = (IEnumerator<Provider>)iterableResult.Value;

            while (iterable.MoveNext()) {
                iterator.SetValue(iterable.Current);
                yield return Call<bool, bool>(InterpretStatement(statement.Body), returnStatus);

                if ((bool)returnStatus)
                {
                    yield return true;
                    yield break;
                }
            }

            iterator.SetValue(Constants.Nil);
            yield return false;
        }
        private IEnumerable<Result<bool>> InterpretWhileStatement(WhileStatement statement)
        {
            Result<ReferenceValue> testValue = new Result<ReferenceValue>();
            Result<bool> returnValue = new Result<bool>();
            while (true)
            {
                yield return Call<bool, ReferenceValue>(InterpretExpression(statement.Test), testValue);
                if (!((ReferenceValue)testValue).IsTrue())
                {
                    yield return false;
                    yield break;
                }
                yield return Call<bool, bool>(InterpretStatement(statement.Body), returnValue);
                if ((bool)returnValue)
                {
                    yield return true;
                    yield break;
                }
            }
        }
        private IEnumerable<Result<bool>> InterpretBlockStatement(BlockStatement statement) {
            Result<bool> returnValue = new Result<bool>();
            foreach (Statement s in statement.Body)
            {
                yield return Call<bool, bool>(InterpretStatement(s), returnValue);
                if ((bool)returnValue) {
                    yield return true;
                    yield break;
                }
            }
            yield return false;
        }
        private IEnumerable<Result<bool>> InterpretIfStatement(IfStatement statement) {
            Result<ReferenceValue> testValue = new Result<ReferenceValue>();
            yield return Call<bool, ReferenceValue>(InterpretExpression(statement.Test), testValue);
            if (((ReferenceValue)testValue).IsTrue()) {
                yield return Return(InterpretStatement(statement.Consequent));
            }
            yield return Return(InterpretStatement(statement.Alternate));
        }
        private IEnumerable<Result<bool>> InterpretReturnStatement(ReturnStatement statement)
        {
            if (statement.Argument != null)
            {
                Result<ReferenceValue> returnValue = new Result<ReferenceValue>();
                yield return Call<bool, ReferenceValue>(InterpretExpression(statement.Argument), returnValue);
                CurrentScope.Result = returnValue;
            }
            else { CurrentScope.Result = Constants.Nil; }

            yield return true;
        }
        private IEnumerable<Result<bool>> InterpretPrintStatement(PrintStatement statement)
        {
            Result<ReferenceValue> printValue = new Result<ReferenceValue>();
            yield return Call<bool, ReferenceValue>(InterpretExpression(statement.Argument), printValue);
            if (PrintDelegate != null) {
                yield return Call<bool>(PrintDelegate(this, printValue));
            }
            yield return false;
        }
        private IEnumerable<Result<bool>> InterpretSwapStatement(SwapStatement statement)
        {
            Result<ReferenceValue> value = new Result<ReferenceValue>();
            ReferenceValue left, temp, right;
            yield return Call<bool, ReferenceValue>(InterpretExpression(statement.Left), value);
            left = value;
            yield return Call<bool, ReferenceValue>(InterpretExpression(statement.Right), value);
            right = value;

            temp = left.Clone();
            left.SetValue(right);
            right.SetValue(temp);

            yield return false;
        }
        private IEnumerable<Result<bool>> InterpretRepeatStatement(RepeatStatement statement)
        {
            Result<bool> returnValue = new Result<bool>();
            Result<ReferenceValue> testValue = new Result<ReferenceValue>();
            while (true)
            {
                yield return Call<bool, bool>(InterpretStatement(statement.Body), returnValue);
                if ((bool)returnValue) { yield return true; }
                yield return Call<bool, ReferenceValue>(InterpretExpression(statement.Test), testValue);
                if (((ReferenceValue)testValue).IsTrue()) { yield return false; }
            }
        }
        private IEnumerable<Result<bool>> InterpretExpressionStatement(ExpressionStatement statement) {
            yield return Call<bool, ReferenceValue>(InterpretExpression(statement.Expression));
            yield return false;
        }

        private IEnumerable<Result<ReferenceValue>> InterpretArrayInitializationExpression(ArrayInitializationExpression expression)
        {
            ArrayProvider ap = new ArrayProvider();
            Result<ReferenceValue> result = new Result<ReferenceValue>();
            for (int i = 0; i < expression.Values.Count; ++i)
            {
                yield return Call<ReferenceValue, ReferenceValue>(InterpretExpression(expression.Values[i]), result);
                ap.Value.Add(i + 1, result);
            }
            ap.Length = expression.Values.Count;
            yield return (ReferenceValue)ap;
        }
        private IEnumerable<Result<ReferenceValue>> InterpretSetInitializationExpression(SetInitializationExpression expression)
        {
            HashSet<Provider> values = new HashSet<Provider>();
            Result<ReferenceValue> result = new Result<ReferenceValue>();
            for (int i = 0; i < expression.Values.Count; ++i)
            {
                yield return Call<ReferenceValue, ReferenceValue>(InterpretExpression(expression.Values[i]), result);
                values.Add(((ReferenceValue)result).Provider);
            }
            yield return (ReferenceValue)new SetProvider(values);
        }
        private IEnumerable<Result<ReferenceValue>> InterpretCallExpression(CallExpression expression)
        {
            Result<ReferenceValue> result = new Result<ReferenceValue>();
            ReferenceValue function;
            ReferenceValue[] args = new ReferenceValue[expression.Arguments.Count];

            yield return Call<ReferenceValue, ReferenceValue>(InterpretExpression(expression.Function), result);
            function = result;

            for (int i = 0; i < args.Length; ++i) {
                yield return Call<ReferenceValue, ReferenceValue>(InterpretExpression(expression.Arguments[i]), result);
                args[i] = result;
            }

            yield return Return(function.Call(this, args));
        }
        private IEnumerable<Result<ReferenceValue>> InterpretGroupExpression(GroupExpression expression) {
            Result<ReferenceValue> result = new Result<ReferenceValue>();
            yield return Call<ReferenceValue, ReferenceValue>(InterpretExpression(expression.Expression), result);

            switch (expression.GroupType) {
                case GroupType.Group:
                    yield return result;
                    break;
                case GroupType.Ceiling:
                    yield return Return(((ReferenceValue)result).Ceiling(this));
                    break;
                case GroupType.Floor:
                    yield return Return(((ReferenceValue)result).Floor(this));
                    break;
            }
        }
        private IEnumerable<Result<ReferenceValue>> InterpretLogicalExpression(BinaryExpression expression)
        {
            Result<ReferenceValue> result = new Result<ReferenceValue>();
            ReferenceValue left;
            yield return Call<ReferenceValue, ReferenceValue>(InterpretExpression(expression.Left), result);
            left = result;

            switch (expression.Operator)
            {
                case BinaryOperator.LogicalAnd:
                    if (left.IsTrue() == false) { yield return (ReferenceValue)Constants.Zero; }
                    break;
                case BinaryOperator.LogicalOr:
                    if (left.IsTrue()) { yield return (ReferenceValue)Constants.One; }
                    break;
            }
            yield return Call<ReferenceValue, ReferenceValue>(InterpretExpression(expression.Right), result);
            yield return ToIntegerConstant(((ReferenceValue)result).IsTrue());
        }
        private IEnumerable<Result<ReferenceValue>> InterpretUnaryExpression(UnaryExpression expression)
        {
            Result<ReferenceValue> result = new Result<ReferenceValue>();
            ReferenceValue argument;
            yield return Call<ReferenceValue, ReferenceValue>(InterpretExpression(expression.Argument), result);
            argument = result;

            switch (expression.Operator)
            {
                case UnaryOperator.LogicalNot:
                    yield return Return(argument.Negate(this));
                    break;
                case UnaryOperator.Opposite:
                    yield return Return(argument.Opposite(this));
                    break;
            }
        }
        private IEnumerable<Result<ReferenceValue>> InterpretBinaryExpression(BinaryExpression expression)
        {
            switch (expression.Operator) {
                case BinaryOperator.LogicalOr:
                case BinaryOperator.LogicalAnd:
                    yield return Return(InterpretLogicalExpression(expression));
                    break;
            }

            Result<ReferenceValue> result = new Result<ReferenceValue>();
            ReferenceValue left;
            ReferenceValue right;
            yield return Call<ReferenceValue, ReferenceValue>(InterpretExpression(expression.Left), result);
            left = result;
            yield return Call<ReferenceValue, ReferenceValue>(InterpretExpression(expression.Right), result);
            right = result;
            Result<bool> intermediate = new Result<bool>();
            switch (expression.Operator)
            {
                case BinaryOperator.Assign:
                    left.SetValue(right);
                    yield return left;
                    break;
                case BinaryOperator.Different:
                    yield return Call<ReferenceValue, bool>(left.IsDifferent(this, right), intermediate);
                    break;
                case BinaryOperator.Equal:
                    yield return Call<ReferenceValue, bool>(left.IsEqual(this, right), intermediate);
                    break;
                case BinaryOperator.Greater:
                    yield return Call<ReferenceValue, bool>(left.IsGreater(this, right), intermediate);
                    break;
                case BinaryOperator.GreaterEqual:
                    yield return Call<ReferenceValue, bool>(left.IsGreaterEqual(this, right), intermediate);
                    break;
                case BinaryOperator.Less:
                    yield return Call<ReferenceValue, bool>(left.IsLess(this, right), intermediate);
                    break;
                case BinaryOperator.LessEqual:
                    yield return Call<ReferenceValue, bool>(left.IsLessEqual(this, right), intermediate);
                    break;
                case BinaryOperator.Div:
                    yield return Return(left.Divide(this, right));
                    break;
                case BinaryOperator.Minus:
                    yield return Return(left.Subtract(this, right));
                    break;
                case BinaryOperator.Mult:
                    yield return Return(left.Multiply(this, right));
                    break;
                case BinaryOperator.Plus:
                    yield return Return(left.Add(this, right));
                    break;
                case BinaryOperator.Remainder:
                    yield return Return(left.Reminder(this, right));
                    break;
                //set operators
                case BinaryOperator.Union:
                    yield return Return(left.Union(this, right));
                    break;
                case BinaryOperator.Intersection:
                    yield return Return(left.Intersect(this, right));
                    break;
                case BinaryOperator.SetMinus:
                    yield return Return(left.SetDifference(this, right));
                    break;
                case BinaryOperator.SubsetOf:
                    yield return Call<ReferenceValue, bool>(left.IsStrictSubsetOf(this, right), intermediate);
                    break;
                case BinaryOperator.SupersetOf:
                    yield return Call<ReferenceValue, bool>(left.IsStrictSupersetOf(this, right), intermediate);
                    break;
                case BinaryOperator.SubsetOfEqual:
                    yield return Call<ReferenceValue, bool>(left.IsSubsetOf(this, right), intermediate);
                    break;
                case BinaryOperator.SupersetOfEqual:
                    yield return Call<ReferenceValue, bool>(left.IsSupersetOf(this, right), intermediate);
                    break;
                case BinaryOperator.ElementOf:
                    yield return Call<ReferenceValue, bool>(left.BelongsTo(this, right), intermediate);
                    break;
                case BinaryOperator.NotElementOf:
                    yield return Call<ReferenceValue, bool>(left.NotBelongsTo(this, right), intermediate);
                    break;
            }
            yield return ToIntegerConstant(intermediate);
        }
        private IEnumerable<Result<ReferenceValue>> InterpretArrayRangeExpression(ArrayRangeExpression expression) {
            Result<ReferenceValue> result = new Result<ReferenceValue>();
            ReferenceValue array, start, end;

            yield return Call<ReferenceValue, ReferenceValue>(InterpretExpression(expression.Array), result);
            array = result;
            yield return Call<ReferenceValue, ReferenceValue>(InterpretExpression(expression.Start), result);
            start = result;
            yield return Call<ReferenceValue, ReferenceValue>(InterpretExpression(expression.End), result);
            end = result;

            yield return Return(array.Slice(this, start, end));
        }
        private IEnumerable<Result<ReferenceValue>> InterpretArrayExpression(ArrayExpression expression)
        {
            Result<ReferenceValue> result = new Result<ReferenceValue>();
            ReferenceValue array, index;
            yield return Call<ReferenceValue, ReferenceValue>(InterpretExpression(expression.Array), result);
            array = result;
            yield return Call<ReferenceValue, ReferenceValue>(InterpretExpression(expression.Index), result);
            index = result;

            yield return array.GetIndexedProperty(index);
        }
        private IEnumerable<Result<ReferenceValue>> InterpretMemberAccessor(MemberAccessor expression)
        {
            Result<ReferenceValue> result = new Result<ReferenceValue>();
            yield return Call<ReferenceValue, ReferenceValue>(InterpretExpression(expression.Object), result);
        
            yield return ((ReferenceValue)result).GetProperty(expression.Member.Name);
        }

        private IEnumerable<Result<ReferenceValue>> InterpretExpression(Expression expression)
        {
            switch (expression.NodeType)
            {
                case NodeType.ArrayInitializationExpression:
                    yield return Return(InterpretArrayInitializationExpression((ArrayInitializationExpression)expression));
                    break;
                case NodeType.SetInitializationExpression:
                    yield return Return(InterpretSetInitializationExpression((SetInitializationExpression)expression));
                    break;
                case NodeType.ArrayExpression:
                    yield return Return(InterpretArrayExpression((ArrayExpression)expression));
                    break;
                case NodeType.CallExpression:
                    yield return Return(InterpretCallExpression((CallExpression)expression));
                    break;
                case NodeType.MemberAccessor:
                    yield return Return(InterpretMemberAccessor((MemberAccessor)expression));
                    break;
                case NodeType.ArrayRangeExpression:
                    yield return Return(InterpretArrayRangeExpression((ArrayRangeExpression)expression));
                    break;
                case NodeType.BinaryExpression:
                    yield return Return(InterpretBinaryExpression((BinaryExpression)expression));
                    break;
                case NodeType.UnaryExpression:
                    yield return Return(InterpretUnaryExpression((UnaryExpression)expression));
                    break;
                case NodeType.GroupExpression:
                    yield return Return(InterpretGroupExpression((GroupExpression)expression));
                    break;
                case NodeType.Identifier:
                    yield return GetVariable(((Identifier)expression).Name);
                    break;
                default:
                    yield return InterpretConstant((Constant)expression);
                    break;
            }
        }

        private ReferenceValue InterpretConstant(Constant expression)
        {
            switch (expression.NodeType)
            {
                case NodeType.ConstantFloat:
                    double dblValue = ((ConstantFloat)expression).Value;
                    if (double.IsNegativeInfinity(dblValue)) { return Constants.NegativeInfinity; }
                    if (double.IsPositiveInfinity(dblValue)) { return Constants.PositiveInfinity; }
                    if (double.IsNaN(dblValue)) { return Constants.NaN; }
                    return new FloatConstantProvider(dblValue);
                case NodeType.ConstantInteger:
                    return new IntegerConstantProvider(((ConstantInteger)expression).Value);
                case NodeType.ConstantInfinity:
                    return Constants.PositiveInfinity;
                case NodeType.ConstantNil:
                    return Constants.Nil;
                case NodeType.ConstantEmptySet:
                    return new SetProvider();
            }
            throw new Exception("Constant unknown");
        }

        private static ReferenceValue ToIntegerConstant(bool value)
        {
            return value ? Constants.One : Constants.Zero;
        }

        public AstInterpreter(ProgramStatement statement) {
            GlobalScope = new Dictionary<string, ReferenceValue>();
            FunctionScope = new Stack<RuntimeMethodInfo>();
            CurrentScope = null;
            Program = statement;

            //E' importante mantenere questo ordine
            foreach (Statement content in statement.Body)
            {
                if (content.NodeType == NodeType.FunctionDeclaration)
                {
                    MethodInfo minfo = new MethodInfo((FunctionDeclaration)content);
                    GlobalScope.Add(minfo.Name, new FunctionProvider(minfo));
                }
            }
            
            string[] globals = MethodInfo.GetVariableUsed(statement);
            foreach (string v in globals) {
                if (!GlobalScope.ContainsKey(v))
                {
#if TYPE_UNDEFINED
                    GlobalScope.Add(v, Constants.Undefined);
#else
                    GlobalScope.Add(v, Constants.Nil);
#endif
                }
            }
            

        }
        public void Interpret() {
            foreach (Statement stat in Program.Body) {
                if (stat.NodeType != NodeType.FunctionDeclaration) {
                    CallImmediate(InterpretStatement(stat));
                }
            }
        }
    }
}
