﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Numerics;
using PseudoInterpreter.Recursive;

namespace PseudoInterpreter.Interpreter
{
    [System.ComponentModel.ImmutableObject(true)]
    public sealed class FloatConstantProvider : Provider, IMathProvider
    {
        public readonly double Value;

        internal FloatConstantProvider(double integer) { Value = integer; }

        #region Provider
        public override IEnumerable<Result<Provider>> Negate(AstInterpreter ai, ReferenceValue thisReference)
        {
            yield return (this.Value == 0 ? Constants.One : Constants.Zero);
        }
        public override bool IsTrue(ReferenceValue thisReference)
        {
            return Value != 0;
        }
        public override IEnumerable<Result<int>> Compare(AstInterpreter ai, ReferenceValue thisReference, ReferenceValue rv)
        {
            Provider p = rv.Provider;
            Type type = p.GetType();
            if (type == typeof(FloatConstantProvider))
            {
                yield return this.Value.CompareTo(((FloatConstantProvider)p).Value);
            }
            if (type == typeof(IntegerConstantProvider))
            {
                //more complex...
                BigInteger thisInt = (BigInteger)this.Value;
                BigInteger other = ((IntegerConstantProvider)p).Value;
                if (thisInt == other)
                {
                    //decimal part exists?
                    double rest = this.Value - Math.Truncate(this.Value);
                    if (rest == 0) { yield return EQUAL; }
                    if (rest < 0) { yield return LESS; }
                    yield return GREATER;
                }
                else { yield return thisInt.CompareTo(other); }
            }
            if (type == typeof(SpecialConstantProvider))
            {
                SpecialConstantProvider a2 = (SpecialConstantProvider)p;
                if (a2.IsNaN) { yield return INCOMPARABLE; }
                if (a2.Sign) { yield return LESS; }
                yield return GREATER;
            }

            yield return INCOMPARABLE;
        }
        #endregion

        #region IMathProvider
        public IEnumerable<Result<Provider>> Add(AstInterpreter ai, ReferenceValue thisReference, ReferenceValue rv)
        {
            Provider a2 = rv.Provider;
            Type type = a2.GetType();
            if (type == typeof(FloatConstantProvider))
            {
                //FloatConstantProvider
                yield return SanitizeDouble(this.Value + ((FloatConstantProvider)a2).Value);
            }
            else if (type == typeof(IntegerConstantProvider))
            {
                //IntegerConstantProvider
                yield return SanitizeDouble(this.Value + (double)((IntegerConstantProvider)a2).Value);
            }
            else if (type == typeof(SpecialConstantProvider))
            {
                //SpecialConstantProvider
                yield return rv.Provider;
            }
            throw new InvalidOperationException();
        }
        public IEnumerable<Result<Provider>> Subtract(AstInterpreter ai, ReferenceValue thisReference, ReferenceValue rv)
        {
            Provider p = rv.Provider;
            Type type = p.GetType();
            if (type == typeof(FloatConstantProvider))
            {
                yield return SanitizeDouble(this.Value - ((FloatConstantProvider)p).Value);
            }
            else if (type == typeof(IntegerConstantProvider))
            {
                yield return SanitizeDouble(this.Value - (double)((IntegerConstantProvider)p).Value);
            }
            if (type == typeof(SpecialConstantProvider))
            {
                SpecialConstantProvider a2 = (SpecialConstantProvider)p;
                if (a2.IsNaN) { yield return rv.Provider; }
                yield return a2.Sign ? Constants.NegativeInfinity : Constants.PositiveInfinity;
            }
            throw new InvalidOperationException();
        }
        public IEnumerable<Result<Provider>> Multiply(AstInterpreter ai, ReferenceValue thisReference, ReferenceValue rv)
        {
            Provider p = rv.Provider;
            Type type = p.GetType();
            if (type == typeof(FloatConstantProvider))
            {
                yield return SanitizeDouble(this.Value * ((FloatConstantProvider)p).Value);
            }
            if (type == typeof(IntegerConstantProvider))
            {
                yield return SanitizeDouble(this.Value * (double)((IntegerConstantProvider)p).Value);
            }
            else if (type == typeof(SpecialConstantProvider))
            {
                SpecialConstantProvider a2 = (SpecialConstantProvider)p;
                if (this.Value == 0 || a2.IsNaN) { yield return Constants.NaN; }
                int sign = (this.Value < 0 ? -1 : 1) * (a2.Sign ? 1 : -1);
                yield return sign < 0 ? Constants.NegativeInfinity : Constants.PositiveInfinity;
            }
            throw new InvalidOperationException();
        }
        public IEnumerable<Result<Provider>> Divide(AstInterpreter ai, ReferenceValue thisReference, ReferenceValue rv)
        {
            Provider p = rv.Provider;
            Type type = p.GetType();
            if (type == typeof(FloatConstantProvider))
            {
                yield return SanitizeDouble(this.Value / ((FloatConstantProvider)p).Value);
            }
            if (type == typeof(IntegerConstantProvider))
            {
                yield return SanitizeDouble(this.Value / (double)((IntegerConstantProvider)p).Value);
            }

            if (type == typeof(SpecialConstantProvider))
            {
                if (((SpecialConstantProvider)p).IsNaN) { yield return p; }
                yield return Constants.Zero;
            }
            throw new InvalidOperationException();
        }
        public IEnumerable<Result<Provider>> Remainder(AstInterpreter ai, ReferenceValue thisReference, ReferenceValue rv)
        {
            Provider p = rv.Provider;
            Type type = p.GetType();
            if (type == typeof(FloatConstantProvider))
            {
                double v2 = ((FloatConstantProvider)p).Value;
                double intermediate = Math.Truncate(this.Value / v2);
                yield return SanitizeDouble(this.Value - intermediate * v2);
            }
            if (type == typeof(IntegerConstantProvider))
            {
                BigInteger intermediate = (BigInteger)this.Value % ((IntegerConstantProvider)p).Value;
                yield return SanitizeDouble((double)intermediate + (this.Value - Math.Truncate(this.Value)));
            }
            if (type == typeof(SpecialConstantProvider))
            {
                yield return Constants.NaN;
            }
            throw new InvalidOperationException();
        }

        public IEnumerable<Result<Provider>> Increment(AstInterpreter ai, ReferenceValue thisReference)
        {
            yield return SanitizeDouble(this.Value + 1);
        }
        public IEnumerable<Result<Provider>> Decrement(AstInterpreter ai, ReferenceValue thisReference)
        {
            yield return SanitizeDouble(this.Value - 1);
        }
        public IEnumerable<Result<Provider>> Opposite(AstInterpreter ai, ReferenceValue thisReference)
        {
            yield return SanitizeDouble(-this.Value);
        }

        public IEnumerable<Result<Provider>> Ceiling(AstInterpreter ai, ReferenceValue thisReference)
        {
            yield return new IntegerConstantProvider((BigInteger)Math.Ceiling(this.Value));
        }
        public IEnumerable<Result<Provider>> Floor(AstInterpreter ai, ReferenceValue thisReference)
        {
            yield return new IntegerConstantProvider((BigInteger)Math.Floor(this.Value));
        }
        #endregion
        
        #region HashSet Related
        public override int GetHashCode()
        {
            if (Math.Truncate(this.Value) == this.Value) {
                return new BigInteger(this.Value).GetHashCode();
            }
            return this.Value.GetHashCode();
        }
        public override bool Equals(object obj)
        {
            Type type = obj.GetType();
            if (type == typeof(FloatConstantProvider))
            {
                return Value.Equals(((FloatConstantProvider)obj).Value);
            }
            if (type == typeof(IntegerConstantProvider))
            {
                if (Math.Truncate(this.Value) != this.Value) { return false; }
                return ((IntegerConstantProvider)obj).Value == (BigInteger)this.Value;
            }
            return false;
        }
        #endregion

        internal static Provider SanitizeDouble(double d) {
            if (double.IsNaN(d)) { return Constants.NaN; }
            if (double.IsNegativeInfinity(d)) { return Constants.NegativeInfinity; }
            if (double.IsPositiveInfinity(d)) { return Constants.PositiveInfinity; }
            return new FloatConstantProvider(d);
        }
    }
}
