﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Numerics;
using PseudoInterpreter.Recursive;

namespace PseudoInterpreter.Interpreter
{
    [System.ComponentModel.ImmutableObject(true)]
    public sealed class IntegerConstantProvider : Provider, IMathProvider
    {
        public readonly BigInteger Value;

        public IntegerConstantProvider(BigInteger integer) { Value = integer; }


        #region Provider
        public override IEnumerable<Result<int>> Compare(AstInterpreter ai, ReferenceValue thisReference, ReferenceValue rv)
        {
            Provider p = rv.Provider;
            Type type = p.GetType();
            if (type == typeof(IntegerConstantProvider))
            {
                yield return this.Value.CompareTo(((IntegerConstantProvider)p).Value);
            }
            if (type == typeof(FloatConstantProvider))
            {
                //more complex...
                double other = ((FloatConstantProvider)p).Value;
                BigInteger otherInt = (BigInteger)other;
                if (this.Value == otherInt)
                {
                    //decimal part exists?
                    double rest = other - Math.Truncate(other);
                    if (rest == 0) { yield return EQUAL; }
                    if (rest < 0) { yield return GREATER; }
                    yield return LESS;
                }
                else { yield return this.Value.CompareTo(otherInt); }
            }
            if (type == typeof(SpecialConstantProvider))
            {
                SpecialConstantProvider a2 = (SpecialConstantProvider)p;
                if (a2.IsNaN) { yield return INCOMPARABLE; }
                if (a2.Sign) { yield return LESS; }
                yield return GREATER;
            }

            yield return INCOMPARABLE;
        }
        public override IEnumerable<Result<Provider>> Negate(AstInterpreter ai, ReferenceValue thisReference)
        {
            yield return this.Value.IsZero ? Constants.One : Constants.Zero;
        }

        public override bool IsTrue(ReferenceValue thisReference)
        {
            return !Value.IsZero;
        }
        #endregion
        #region IMathProvider
        public IEnumerable<Result<Provider>> Add(AstInterpreter ai, ReferenceValue thisReference, ReferenceValue rv)
        {
            Provider a2 = rv.Provider;
            Type type = a2.GetType();
            if (type == typeof(FloatConstantProvider))
            {
                //FloatConstantProvider
                yield return FloatConstantProvider.SanitizeDouble((double)this.Value + ((FloatConstantProvider)a2).Value);
            }
            if (type == typeof(IntegerConstantProvider))
            {
                //IntegerConstantProvider
                yield return new IntegerConstantProvider(this.Value + ((IntegerConstantProvider)a2).Value);
            }
            else if (type == typeof(SpecialConstantProvider))
            {
                //SpecialConstantProvider
                yield return rv.Provider;
            }

            throw new InvalidOperationException();
        }
        public IEnumerable<Result<Provider>> Subtract(AstInterpreter ai, ReferenceValue thisReference, ReferenceValue rv)
        {
            Provider p = rv.Provider;
            Type type = p.GetType();
            if (type == typeof(IntegerConstantProvider))
            {
                yield return new IntegerConstantProvider(this.Value - ((IntegerConstantProvider)p).Value);
            }
            if (type == typeof(FloatConstantProvider))
            {
                yield return FloatConstantProvider.SanitizeDouble((double)this.Value - ((FloatConstantProvider)p).Value);
            }
            if (type == typeof(SpecialConstantProvider))
            {
                SpecialConstantProvider a2 = (SpecialConstantProvider)p;
                if (a2.IsNaN) { yield return rv.Provider; }
                yield return a2.Sign ? Constants.NegativeInfinity : Constants.PositiveInfinity;
            }

            throw new InvalidOperationException();
        }
        public IEnumerable<Result<Provider>> Multiply(AstInterpreter ai, ReferenceValue thisReference, ReferenceValue rv)
        {
            Provider p = rv.Provider;
            Type type = p.GetType();
            if (type == typeof(IntegerConstantProvider))
            {
                yield return new IntegerConstantProvider(this.Value * ((IntegerConstantProvider)p).Value);
            }
            if (type == typeof(FloatConstantProvider))
            {
                yield return FloatConstantProvider.SanitizeDouble((double)this.Value * ((FloatConstantProvider)p).Value);
            }
            if (type == typeof(SpecialConstantProvider))
            {
                SpecialConstantProvider a2 = (SpecialConstantProvider)p;
                if (this.Value.IsZero || a2.IsNaN) { yield return Constants.NaN; }
                int sign = this.Value.Sign * (a2.Sign ? 1 : -1);
                yield return sign < 0 ? Constants.NegativeInfinity : Constants.PositiveInfinity;
            }

            throw new InvalidOperationException();
        }
        public IEnumerable<Result<Provider>> Divide(AstInterpreter ai, ReferenceValue thisReference, ReferenceValue rv)
        {
            Provider p = rv.Provider;
            Type type = p.GetType();
            if (type == typeof(IntegerConstantProvider))
            {
                BigInteger other = ((IntegerConstantProvider)p).Value;
                if (other.IsZero) {
                    //special treatment for division by zero
                    if (this.Value.IsZero) { yield return Constants.NaN; }
                    if (this.Value > 0) { yield return Constants.PositiveInfinity; }
                    yield return Constants.NegativeInfinity;
                    yield break;
                }
                yield return new IntegerConstantProvider(this.Value / other);
            }
            if (type == typeof(FloatConstantProvider))
            {
                yield return FloatConstantProvider.SanitizeDouble((double)this.Value / ((FloatConstantProvider)p).Value);
            }
            if (type == typeof(SpecialConstantProvider))
            {
                if (((SpecialConstantProvider)p).IsNaN) { yield return p; }
                yield return Constants.Zero;
            }

            throw new InvalidOperationException();
        }
        public IEnumerable<Result<Provider>> Remainder(AstInterpreter ai, ReferenceValue thisReference, ReferenceValue rv)
        {
            Provider p = rv.Provider;
            Type type = p.GetType();
            if (type == typeof(IntegerConstantProvider))
            {
                BigInteger other = ((IntegerConstantProvider)p).Value;
                if (other.IsZero) { yield return Constants.NaN; }
                yield return new IntegerConstantProvider(this.Value % other);
            }
            if (type == typeof(FloatConstantProvider))
            {
                double v2 = ((FloatConstantProvider)p).Value;
                double intermediate = Math.Truncate((double)this.Value / v2);
                yield return FloatConstantProvider.SanitizeDouble((double)this.Value - intermediate * v2);
            }
            if (type == typeof(SpecialConstantProvider))
            {
                yield return Constants.NaN;
            }

            throw new InvalidOperationException();
        }
        
        public IEnumerable<Result<Provider>> Increment(AstInterpreter ai, ReferenceValue thisReference)
        {
            yield return new IntegerConstantProvider(this.Value + 1);
        }
        public IEnumerable<Result<Provider>> Decrement(AstInterpreter ai, ReferenceValue thisReference)
        {
            yield return new IntegerConstantProvider(this.Value - 1);
        }
        public IEnumerable<Result<Provider>> Opposite(AstInterpreter ai, ReferenceValue thisReference)
        {
            yield return new IntegerConstantProvider(-this.Value);
        }
        
        public IEnumerable<Result<Provider>> Ceiling(AstInterpreter ai, ReferenceValue thisReference)
        {
            yield return this;
        }
        public IEnumerable<Result<Provider>> Floor(AstInterpreter ai, ReferenceValue thisReference)
        {
            yield return this;
        }
        #endregion

        #region HashSet Related
        public override int GetHashCode()
        {
            return Value.GetHashCode();
        }
        public override bool Equals(object obj)
        {
            Type type = obj.GetType();
            if (type == typeof(IntegerConstantProvider))
            {
                return Value.Equals(((IntegerConstantProvider)obj).Value);
            }
            if (type == typeof(FloatConstantProvider)) {
                double v2 = ((FloatConstantProvider)obj).Value;
                if (Math.Truncate(v2) != v2) { return false; }
                return Value == (BigInteger)v2;
            }
            return false;
        }
        #endregion

        public static implicit operator IntegerConstantProvider(long value)
        {
            if (value == 0) { return Constants.Zero; }
            else if (value == 1) { return Constants.One; }
            return new IntegerConstantProvider(value);
        }
        public static implicit operator IntegerConstantProvider(int value)
        {
            if (value == 0) { return Constants.Zero; }
            else if (value == 1) { return Constants.One; }
            return new IntegerConstantProvider(value);
        }
        public static implicit operator IntegerConstantProvider(BigInteger value)
        {
            if (value == 0) { return Constants.Zero; }
            else if (value == 1) { return Constants.One; }
            return new IntegerConstantProvider(value);
        }
        public static implicit operator IntegerConstantProvider(bool value)
        {
            return value ? Constants.One : Constants.Zero;
        }
    }
}
