﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Collections.ObjectModel;
using PseudoInterpreter.Ast;

namespace PseudoInterpreter.Interpreter
{
    [System.ComponentModel.ImmutableObject(true)]
    public sealed class MethodInfo
    {
        public readonly string Name;
        public readonly ReadOnlyCollection<string> Parameters;
        public readonly ReadOnlyCollection<string> Locals;
        public readonly FunctionDeclaration Declaration;

        //questo metodo dovrebbe essere spostato
        internal static string[] GetVariableUsed(Node root) {

            HashSet<string> variableFound = new HashSet<string>();
            Stack<Node> nodeToBeVisited = new Stack<Node>();
            List<string> localVariables = new List<string>();

            nodeToBeVisited.Push(root);
            while (nodeToBeVisited.Count != 0)
            {
                Node node = nodeToBeVisited.Pop();
                if (node == null) { continue; }
                switch (node.NodeType)
                {
                    case NodeType.ArrayExpression:
                        nodeToBeVisited.Push(((ArrayExpression)node).Array);
                        nodeToBeVisited.Push(((ArrayExpression)node).Index);
                        break;
                    case NodeType.ArrayRangeExpression:
                        nodeToBeVisited.Push(((ArrayRangeExpression)node).Array);
                        nodeToBeVisited.Push(((ArrayRangeExpression)node).Start);
                        nodeToBeVisited.Push(((ArrayRangeExpression)node).End);
                        break;
                    case NodeType.BinaryExpression:
                        nodeToBeVisited.Push(((BinaryExpression)node).Left);
                        nodeToBeVisited.Push(((BinaryExpression)node).Right);
                        break;
                    case NodeType.BlockStatement:
                        BlockStatement blockStatement = (BlockStatement)node;
                        for (int i = 0; i < blockStatement.Body.Count; ++i)
                        {
                            nodeToBeVisited.Push(blockStatement.Body[i]);
                        }
                        break;
                    case NodeType.ExpressionStatement:
                        nodeToBeVisited.Push(((ExpressionStatement)node).Expression);
                        break;
                    case NodeType.ForStatement:
                        nodeToBeVisited.Push(((ForStatement)node).Initial);
                        nodeToBeVisited.Push(((ForStatement)node).End);
                        nodeToBeVisited.Push(((ForStatement)node).Body);
                        break;
                    case NodeType.CallExpression:
                        CallExpression callExpression = (CallExpression)node;
                        nodeToBeVisited.Push(callExpression.Function);
                        for (int i = 0; i < callExpression.Arguments.Count; ++i)
                        {
                            nodeToBeVisited.Push(callExpression.Arguments[i]);
                        }
                        break;
                    case NodeType.GroupExpression:
                        nodeToBeVisited.Push(((GroupExpression)node).Expression);
                        break;
                    case NodeType.IfStatement:
                        nodeToBeVisited.Push(((IfStatement)node).Test);
                        nodeToBeVisited.Push(((IfStatement)node).Consequent);
                        nodeToBeVisited.Push(((IfStatement)node).Alternate);
                        break;
                    case NodeType.MemberAccessor:
                        nodeToBeVisited.Push(((MemberAccessor)node).Object);
                        break;
                    case NodeType.ReturnStatement:
                        nodeToBeVisited.Push(((ReturnStatement)node).Argument);
                        break;
                    case NodeType.UnaryExpression:
                        nodeToBeVisited.Push(((UnaryExpression)node).Argument);
                        break;
                    case NodeType.WhileStatement:
                        nodeToBeVisited.Push(((WhileStatement)node).Test);
                        nodeToBeVisited.Push(((WhileStatement)node).Body);
                        break;
                    case NodeType.Identifier:
                        //important
                        string variableName = ((Identifier)node).Name;
                        if (!variableFound.Contains(variableName))
                        {
                            variableFound.Add(variableName);
                            localVariables.Add(variableName);
                        }
                        break;
                    case NodeType.ProgramStatement:
                        ProgramStatement programStatement = (ProgramStatement)node;
                        foreach (Statement content in programStatement.Body) {
                            if (content.NodeType == NodeType.FunctionDeclaration) { continue; }
                            nodeToBeVisited.Push(content);
                        }
                        break;
                    case NodeType.ArrayInitializationExpression:
                        ArrayInitializationExpression arrayInit = (ArrayInitializationExpression)node;
                        foreach (Expression expr in arrayInit.Values) {
                            nodeToBeVisited.Push(expr);
                        }
                        break;
                    case NodeType.SetInitializationExpression:
                        SetInitializationExpression setInit = (SetInitializationExpression)node;
                        foreach (Expression expr in setInit.Values)
                        {
                            nodeToBeVisited.Push(expr);
                        }
                        break;
                    case NodeType.ForEachStatement:
                        nodeToBeVisited.Push(((ForEachStatement)node).IterationVariable);
                        nodeToBeVisited.Push(((ForEachStatement)node).Iterable);
                        nodeToBeVisited.Push(((ForEachStatement)node).Body);
                        break;
                    case NodeType.PrintStatement:
                        nodeToBeVisited.Push(((PrintStatement)node).Argument);
                        break;
                    case NodeType.RepeatStatement:
                        nodeToBeVisited.Push(((RepeatStatement)node).Test);
                        nodeToBeVisited.Push(((RepeatStatement)node).Body);
                        break;
                    case NodeType.SwapStatement:
                        nodeToBeVisited.Push(((SwapStatement)node).Left);
                        nodeToBeVisited.Push(((SwapStatement)node).Right);
                        break;
                }
            }
            return localVariables.ToArray();
        }
        public MethodInfo(FunctionDeclaration statement) {
            this.Declaration = statement;
            //si occupa di ottenere tutte le variabili (nomi in una tabella), il numero di argomenti ed il loro nome
            HashSet<string> variableFound = new HashSet<string>();
            this.Name = statement.Name.Name;
            string [] parameters = new string[statement.Parameters.Count];
            for (int i = 0; i < parameters.Length; ++i)
            {
                parameters[i] = statement.Parameters[i].Name;
                variableFound.Add(parameters[i]);
            }
            string[] bodyVariables = GetVariableUsed(statement.Body);
            List<string> localVariables = new List<string>();
            for (int i = 0; i < bodyVariables.Length; ++i) {
                if (variableFound.Contains(bodyVariables[i])) { continue; }
                localVariables.Add(bodyVariables[i]);
            }
            this.Locals = localVariables.AsReadOnly();
            this.Parameters = Array.AsReadOnly<string>(parameters);
        }
    }
}
