﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using PseudoInterpreter.Recursive;

namespace PseudoInterpreter.Interpreter
{

    public sealed class ReferenceValue
    {
        public Provider Provider { get; internal set; }
        internal event EventHandler<ValueChangedArgs> ValueChanged;

        public void SetValue(ReferenceValue value)
        {
            SetValue(value.Provider, true);
        }
        public void SetValue(Provider value) {
            SetValue(value, true);
        }
        public void SetValue(Provider value, bool notifyChange)
        {
            this.Provider = value;
#if TYPE_UNDEFINED
            if (this.Provider != null && this.Provider != Constants.Undefined)
#else
            if (this.Provider != null && this.Provider != Constants.Nil)
#endif
            {
                if (notifyChange && this.ValueChanged != null)
                { this.ValueChanged(this, new ValueChangedArgs(this.Provider, value)); }
            }
        }


        public IEnumerable<Result<IEnumerator<Provider>>> GetEnumerator(AstInterpreter ai) {
            yield return ai.Return(((ISetProvider)this.Provider).GetEnumerator(ai, this));
        }
        public ReferenceValue GetProperty(string name)
        {
            return ((IObjectProvider)this.Provider).GetProperty(this, name);
        }
        public ReferenceValue GetIndexedProperty(ReferenceValue index)
        {
            return ((IArrayProvider)this.Provider).GetIndexedProperty(this, index);
        }

        public IEnumerable<Result<bool>> IsEqual(AstInterpreter ai, ReferenceValue rv)
        {
            Result<int> result = new Result<int>();
            yield return ai.Call<bool, int>(this.Provider.Compare(ai, this, rv), result);
            yield return result == 0;
        }
        public IEnumerable<Result<bool>> IsDifferent(AstInterpreter ai, ReferenceValue rv)
        {
            Result<int> result = new Result<int>();
            yield return ai.Call<bool, int>(this.Provider.Compare(ai, this, rv), result);
            yield return result != 0;
        }
        public IEnumerable<Result<bool>> IsGreater(AstInterpreter ai, ReferenceValue rv)
        {
            Result<int> result = new Result<int>();
            yield return ai.Call<bool, int>(this.Provider.Compare(ai, this, rv), result);
            yield return result == 1;
        }
        public IEnumerable<Result<bool>> IsGreaterEqual(AstInterpreter ai, ReferenceValue rv)
        {
            Result<int> result = new Result<int>();
            yield return ai.Call<bool, int>(this.Provider.Compare(ai, this, rv), result);
            int r = result;
            yield return r == 0 || r == 1;
        }
        public IEnumerable<Result<bool>> IsLess(AstInterpreter ai, ReferenceValue rv)
        {
            Result<int> result = new Result<int>();
            yield return ai.Call<bool, int>(this.Provider.Compare(ai, this, rv), result);
            yield return result == -1;
        }
        public IEnumerable<Result<bool>> IsLessEqual(AstInterpreter ai, ReferenceValue rv)
        {
            Result<int> result = new Result<int>();
            yield return ai.Call<bool, int>(this.Provider.Compare(ai, this, rv), result);
            int r = result;
            yield return r == 0 || r == -1;
        }

        public IEnumerable<Result<ReferenceValue>> Multiply(AstInterpreter ai, ReferenceValue rv)
        {
            Result<Provider> result = new Result<Provider>();
            yield return ai.Call<ReferenceValue, Provider>(((IMathProvider)this.Provider).Multiply(ai, this, rv), result);
            yield return new ReferenceValue(result);
        }
        public IEnumerable<Result<ReferenceValue>> Add(AstInterpreter ai, ReferenceValue rv)
        {
            Result<Provider> result = new Result<Provider>();
            yield return ai.Call<ReferenceValue, Provider>(((IMathProvider)this.Provider).Add(ai, this, rv), result);
            yield return new ReferenceValue(result);
        }
        public IEnumerable<Result<ReferenceValue>> Subtract(AstInterpreter ai, ReferenceValue rv)
        {
            Result<Provider> result = new Result<Provider>();
            yield return ai.Call<ReferenceValue, Provider>(((IMathProvider)this.Provider).Subtract(ai, this, rv), result);
            yield return new ReferenceValue(result);
        }
        public IEnumerable<Result<ReferenceValue>> Divide(AstInterpreter ai, ReferenceValue rv)
        {
            Result<Provider> result = new Result<Provider>();
            yield return ai.Call<ReferenceValue, Provider>(((IMathProvider)this.Provider).Divide(ai, this, rv), result);
            yield return new ReferenceValue(result);
        }
        public IEnumerable<Result<ReferenceValue>> Reminder(AstInterpreter ai, ReferenceValue rv)
        {
            Result<Provider> result = new Result<Provider>();
            yield return ai.Call<ReferenceValue, Provider>(((IMathProvider)this.Provider).Remainder(ai, this, rv), result);
            yield return new ReferenceValue(result);
        }
        public IEnumerable<Result<ReferenceValue>> Call(AstInterpreter interpreter, ReferenceValue[] arguments)
        {
            Result<Provider> result = new Result<Provider>();
            yield return interpreter.Call<ReferenceValue, Provider>(((ICallProvider)this.Provider).Call(interpreter, this, arguments), result);
            yield return new ReferenceValue(result);
        }
        public IEnumerable<Result<ReferenceValue>> SetDifference(AstInterpreter ai, ReferenceValue rv)
        {
            Result<Provider> result = new Result<Provider>();
            yield return ai.Call<ReferenceValue, Provider>(((ISetProvider)this.Provider).SetDifference(ai, this, rv), result);
            yield return new ReferenceValue(result);
        }
        public IEnumerable<Result<ReferenceValue>> Union(AstInterpreter ai, ReferenceValue rv)
        {
            Result<Provider> result = new Result<Provider>();
            yield return ai.Call<ReferenceValue, Provider>(((ISetProvider)this.Provider).Union(ai, this, rv), result);
            yield return new ReferenceValue(result);
        }
        public IEnumerable<Result<ReferenceValue>> Intersect(AstInterpreter ai, ReferenceValue rv)
        {
            Result<Provider> result = new Result<Provider>();
            yield return ai.Call<ReferenceValue, Provider>(((ISetProvider)this.Provider).Intersect(ai, this, rv), result);
            yield return new ReferenceValue(result);
        }

        public IEnumerable<Result<bool>> BelongsTo(AstInterpreter ai, ReferenceValue rv)
        {
            yield return ai.Return(((ISetProvider)rv.Provider).Contains(ai, rv, this));
        }
        public IEnumerable<Result<bool>> NotBelongsTo(AstInterpreter ai, ReferenceValue rv)
        {
            Result<bool> result = new Result<bool>();
            yield return ai.Call<bool, bool>(((ISetProvider)rv.Provider).Contains(ai, rv, this), result);
            yield return !(bool)result;
        }
        public IEnumerable<Result<bool>> IsStrictSubsetOf(AstInterpreter ai, ReferenceValue rv)
        {
            Result<int> result = new Result<int>();
            yield return ai.Call<bool, int>(((ISetProvider)this.Provider).SetCompare(ai, this, rv), result);
            yield return result == -1;
        }
        public IEnumerable<Result<bool>> IsStrictSupersetOf(AstInterpreter ai, ReferenceValue rv)
        {
            Result<int> result = new Result<int>();
            yield return ai.Call<bool, int>(((ISetProvider)this.Provider).SetCompare(ai, this, rv), result);
            yield return result == 1;
        }
        public IEnumerable<Result<bool>> IsSubsetOf(AstInterpreter ai, ReferenceValue rv)
        {
            Result<int> result = new Result<int>();
            yield return ai.Call<bool, int>(((ISetProvider)this.Provider).SetCompare(ai, this, rv), result);
            int r = result;
            yield return r == -1 || r == 0;
        }
        public IEnumerable<Result<bool>> IsSupersetOf(AstInterpreter ai, ReferenceValue rv)
        {
            Result<int> result = new Result<int>();
            yield return ai.Call<bool, int>(((ISetProvider)this.Provider).SetCompare(ai, this, rv), result);
            int r = result;
            yield return r == 1 || r == 0;
        }

        public IEnumerable<Result<ReferenceValue>> Negate(AstInterpreter ai)
        {
            Result<Provider> result = new Result<Provider>();
            yield return ai.Call<ReferenceValue, Provider>(this.Provider.Negate(ai, this), result);
            yield return new ReferenceValue(result);
        }
        public IEnumerable<Result<ReferenceValue>> Opposite(AstInterpreter ai)
        {
            Result<Provider> result = new Result<Provider>();
            yield return ai.Call<ReferenceValue, Provider>(((IMathProvider)this.Provider).Opposite(ai, this), result);
            yield return new ReferenceValue(result);
        }
        public IEnumerable<Result<ReferenceValue>> Increment(AstInterpreter ai)
        {
            Result<Provider> result = new Result<Provider>();
            yield return ai.Call<ReferenceValue, Provider>(((IMathProvider)this.Provider).Increment(ai, this), result);
            yield return new ReferenceValue(result);
        }
        public IEnumerable<Result<ReferenceValue>> Decrement(AstInterpreter ai)
        {
            Result<Provider> result = new Result<Provider>();
            yield return ai.Call<ReferenceValue, Provider>(((IMathProvider)this.Provider).Decrement(ai, this), result);
            yield return new ReferenceValue(result);
        }
        public IEnumerable<Result<ReferenceValue>> Ceiling(AstInterpreter ai)
        {
            Result<Provider> result = new Result<Provider>();
            yield return ai.Call<ReferenceValue, Provider>(((IMathProvider)this.Provider).Ceiling(ai, this), result);
            yield return new ReferenceValue(result);
        }
        public IEnumerable<Result<ReferenceValue>> Floor(AstInterpreter ai)
        {
            Result<Provider> result = new Result<Provider>();
            yield return ai.Call<ReferenceValue, Provider>(((IMathProvider)this.Provider).Floor(ai, this), result);
            yield return new ReferenceValue(result);
        }
        public IEnumerable<Result<ReferenceValue>> Slice(AstInterpreter ai, ReferenceValue start, ReferenceValue end)
        {
            Result<Provider> result = new Result<Provider>();
            yield return ai.Call<ReferenceValue, Provider>(((IArrayProvider)this.Provider).Slice(ai, this, start, end), result);
            yield return new ReferenceValue(result);
        }

        
        public bool IsTrue()
        {
            return this.Provider.IsTrue(this);
        }

        public ReferenceValue(Provider provider)
        {
            this.SetValue(provider);
        }
        public ReferenceValue Clone()
        {
            return new ReferenceValue(this.Provider);
        }

        public static implicit operator ReferenceValue(Provider provider)
        {
            return new ReferenceValue(provider);
        }
    }
}
