﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Numerics;
using PseudoInterpreter.Recursive;

namespace PseudoInterpreter.Interpreter
{
    public sealed class SetProvider : Provider, ISetProvider
    {
        public readonly ReadOnlySet<Provider> Value;

        public readonly Dictionary<string, ReferenceValue> Property = new Dictionary<string, ReferenceValue>();

        public SetProvider()
        {
            this.Value = new ReadOnlySet<Provider>(new HashSet<Provider>());
        }
        internal SetProvider(IEnumerable<Provider> set)
        {
            this.Value = new ReadOnlySet<Provider>(new HashSet<Provider>(set));
        }
        internal SetProvider(ISet<Provider> set) {
            this.Value = new ReadOnlySet<Provider>(set);
        }

        #region ISetProvider
        public IEnumerable<Result<bool>> Contains(AstInterpreter ai, ReferenceValue thisReference, ReferenceValue rv)
        {
            yield return Value.Contains(rv.Provider);
        }
        public IEnumerable<Result<Provider>> SetDifference(AstInterpreter ai, ReferenceValue thisReference, ReferenceValue rv)
        {
            Provider p = rv.Provider;
            Type type = p.GetType();
            IEnumerable<Provider> second;
            if (type == typeof(SetProvider)) {
                second = ((SetProvider)p).Value;
            }
            else if (type == typeof(ArrayProvider)) {
                second = ((ArrayProvider)p).Value.Select((x) => x.Value.Provider);
            }
            else { throw new InvalidOperationException(); }

            HashSet<Provider> value = new HashSet<Provider>(this.Value);
            value.ExceptWith(second);
            yield return new SetProvider(value);
        }
        public IEnumerable<Result<Provider>> Intersect(AstInterpreter ai, ReferenceValue thisReference, ReferenceValue rv)
        {
            Provider p = rv.Provider;
            Type type = p.GetType();
            IEnumerable<Provider> second;
            if (type == typeof(SetProvider))
            {
                second = ((SetProvider)p).Value;
            }
            else if (type == typeof(ArrayProvider))
            {
                second = ((ArrayProvider)p).Value.Select((x) => x.Value.Provider);
            }
            else { throw new InvalidOperationException(); }

            HashSet<Provider> value = new HashSet<Provider>(this.Value);
            value.IntersectWith(second);
            yield return new SetProvider(value);
        }
        public IEnumerable<Result<Provider>> Union(AstInterpreter ai, ReferenceValue thisReference, ReferenceValue rv)
        {
            Provider p = rv.Provider;
            Type type = p.GetType();
            IEnumerable<Provider> second;
            if (type == typeof(SetProvider))
            {
                second = ((SetProvider)p).Value;
            }
            else if (type == typeof(ArrayProvider))
            {
                second = ((ArrayProvider)p).Value.Select((x) => x.Value.Provider);
            }
            else { throw new InvalidOperationException(); }

            HashSet<Provider> value = new HashSet<Provider>(this.Value);
            value.UnionWith(second);
            yield return new SetProvider(value);
        }
        public IEnumerable<Result<int>> SetCompare(AstInterpreter ai, ReferenceValue thisReference, ReferenceValue rv)
        {
            Provider p = rv.Provider;
            Type type = p.GetType();
            IEnumerable<Provider> second;
            if (type == typeof(SetProvider))
            {
                second = ((SetProvider)p).Value;
            }
            else if (type == typeof(ArrayProvider))
            {
                second = ((ArrayProvider)p).Value.Select((x) => x.Value.Provider);
            }
            else { throw new InvalidOperationException(); }

            if (this.Value.IsProperSubsetOf(second)) { yield return SUBSET; }
            if (this.Value.IsProperSupersetOf(second)) { yield return SUPERSET; }
            if (this.Value.SetEquals(second)) { yield return EQUAL; }

            yield return INCOMPARABLE;
        }
        public IEnumerable<Result<IEnumerator<Provider>>> GetEnumerator(AstInterpreter ai, ReferenceValue thisReference)
        {
            yield return new Result<IEnumerator<Provider>>(this.Value.GetEnumerator());
        }
        #endregion
        #region IObjectProvider
        public ReferenceValue GetProperty(ReferenceValue thisReference, string name)
        {
            if (!Property.ContainsKey(name)) {
#if TYPE_UNDEFINED
                Property[name] = Constants.Undefined;
#else
                Property[name] = Constants.Nil;
#endif
            }
            return Property[name];
        }
        #endregion

        #region Provider
        public override bool IsTrue(ReferenceValue thisReference)
        {
            return this.Value.Count > 0;
        }
        public override IEnumerable<Result<Provider>> Negate(AstInterpreter ai, ReferenceValue thisReference)
        {
            yield return (this.Value.Count == 0) ? Constants.One : Constants.Zero;
        }
        #endregion
    }
}
