﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using PseudoInterpreter.Recursive;

namespace PseudoInterpreter.Interpreter
{
    [System.ComponentModel.ImmutableObject(true)]
    internal sealed class SpecialConstantProvider : Provider, IMathProvider
    {
        internal readonly bool Sign;
        internal readonly bool IsNaN;

        internal SpecialConstantProvider(bool isNaN, bool sign)
        {
            this.Sign = sign;
            this.IsNaN = isNaN;
        }

        #region Provider
        public override IEnumerable<Result<int>> Compare(AstInterpreter ai, ReferenceValue thisReference, ReferenceValue rv)
        {
            Provider p = rv.Provider;
            Type type = p.GetType();
            if (type == typeof(IntegerConstantProvider) || type == typeof(FloatConstantProvider))
            {
                if (this.IsNaN) { yield return INCOMPARABLE; }
                if (this.Sign) { yield return GREATER; }
                yield return LESS;
            }
            if (type == typeof(SpecialConstantProvider))
            {
                SpecialConstantProvider a2 = (SpecialConstantProvider)p;
                if (this.IsNaN || a2.IsNaN) { yield return INCOMPARABLE; }
                if (this.Sign == a2.Sign) { yield return EQUAL; }
                if (this.Sign) { yield return GREATER; }
                yield return LESS;
            }

            yield return INCOMPARABLE;
        }

        public override IEnumerable<Result<Provider>> Negate(AstInterpreter ai, ReferenceValue thisReference)
        {
            yield return this.IsNaN ? Constants.One : Constants.Zero;
        }
        public override bool IsTrue(ReferenceValue thisReference)
        {
            return !this.IsNaN;
        }
        #endregion

        #region IMathProvider
        public IEnumerable<Result<Provider>> Add(AstInterpreter ai, ReferenceValue thisReference, ReferenceValue rv)
        {
            Provider p = rv.Provider;
            Type type = p.GetType();
            if (type == typeof(IntegerConstantProvider) || type == typeof(FloatConstantProvider))
            {
                yield return this;
            }
            if (type == typeof(SpecialConstantProvider))
            {
                SpecialConstantProvider a2 = (SpecialConstantProvider)p;
                if (this.IsNaN || a2.IsNaN) { yield return Constants.NaN; }
                if (this.Sign != a2.Sign) { yield return Constants.NaN; }
                yield return this;
            }

            throw new InvalidOperationException();
        }
        public IEnumerable<Result<Provider>> Subtract(AstInterpreter ai, ReferenceValue thisReference, ReferenceValue rv)
        {
            Provider p = rv.Provider;
            Type type = p.GetType();
            if (type == typeof(IntegerConstantProvider) || type == typeof(FloatConstantProvider))
            {
                yield return this;
            }
            if (type == typeof(SpecialConstantProvider))
            {
                SpecialConstantProvider a2 = (SpecialConstantProvider)p;
                if (this.IsNaN || a2.IsNaN) { yield return Constants.NaN; }
                if (this.Sign == a2.Sign) { yield return Constants.NaN; }
                yield return this;
            }

            throw new InvalidOperationException();
        }
        public IEnumerable<Result<Provider>> Multiply(AstInterpreter ai, ReferenceValue thisReference, ReferenceValue rv)
        {
            Provider p = rv.Provider;
            Type type = p.GetType();
            if (type == typeof(IntegerConstantProvider))
            {
                IntegerConstantProvider a2 = (IntegerConstantProvider)p;
                if (a2.Value.IsZero || this.IsNaN) { yield return Constants.NaN; }
                int sign = a2.Value.Sign * (this.Sign ? 1 : -1);
                yield return sign < 0 ? Constants.NegativeInfinity : Constants.PositiveInfinity;
            }
            if (type == typeof(FloatConstantProvider))
            {
                FloatConstantProvider a2 = (FloatConstantProvider)p;
                if (a2.Value == 0 || this.IsNaN) { yield return Constants.NaN; }
                int sign = (a2.Value < 0 ? -1 : 1) * (this.Sign ? 1 : -1);
                yield return sign < 0 ? Constants.NegativeInfinity : Constants.PositiveInfinity;
            }
            if (type == typeof(SpecialConstantProvider))
            {
                SpecialConstantProvider a2 = (SpecialConstantProvider)p;
                if (this.IsNaN || a2.IsNaN) { yield return Constants.NaN; }
                yield return this.Sign == a2.Sign ? Constants.PositiveInfinity : Constants.NegativeInfinity;
            }

            throw new InvalidOperationException();
        }
        public IEnumerable<Result<Provider>> Divide(AstInterpreter ai, ReferenceValue thisReference, ReferenceValue rv)
        {
            Provider p = rv.Provider;
            Type type = p.GetType();
            if (type == typeof(IntegerConstantProvider))
            {
                IntegerConstantProvider a2 = (IntegerConstantProvider)p;
                if (a2.Value.IsZero || this.IsNaN) { yield return this; }
                int sign = a2.Value.Sign * (this.Sign ? 1 : -1);
                yield return sign < 0 ? Constants.NegativeInfinity : Constants.PositiveInfinity;
            }
            if (type == typeof(FloatConstantProvider))
            {
                FloatConstantProvider a2 = (FloatConstantProvider)p;
                if (a2.Value == 0 || this.IsNaN) { yield return this; }
                int sign = (a2.Value < 0 ? -1 : 1) * (this.Sign ? 1 : -1);
                yield return sign < 0 ? Constants.NegativeInfinity : Constants.PositiveInfinity;
            }
            if (type == typeof(SpecialConstantProvider))
            {
                yield return Constants.NaN;
            }

            throw new InvalidOperationException();
        }
        public IEnumerable<Result<Provider>> Remainder(AstInterpreter ai, ReferenceValue thisReference, ReferenceValue rv)
        {
            Type type = rv.Provider.GetType();
            if (type == typeof(IntegerConstantProvider) || type == typeof(FloatConstantProvider) || type == typeof(SpecialConstantProvider))
            {
                yield return Constants.NaN;
            }

            throw new InvalidOperationException();
        }

        public IEnumerable<Result<Provider>> Increment(AstInterpreter ai, ReferenceValue thisReference)
        {
            yield return this;
        }
        public IEnumerable<Result<Provider>> Decrement(AstInterpreter ai, ReferenceValue thisReference)
        {
            yield return this;
        }
        public IEnumerable<Result<Provider>> Opposite(AstInterpreter ai, ReferenceValue thisReference)
        {
            if (this.IsNaN) { yield return this; }
            yield return this.Sign ? Constants.NegativeInfinity : Constants.PositiveInfinity;
        }

        public IEnumerable<Result<Provider>> Ceiling(AstInterpreter ai, ReferenceValue thisReference)
        {
            yield return this;
        }
        public IEnumerable<Result<Provider>> Floor(AstInterpreter ai, ReferenceValue thisReference)
        {
            yield return this;
        }

        #endregion
    }
}
