﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Numerics;
using System.Threading.Tasks;

namespace PseudoInterpreter.Tokenizer
{
    public sealed class Tokenizer
    {
        private TokenizerReader Reader;
        private List<int> IndentationLevels;
        private List<Token> EmittedTokens;
        private int CurrentIndentation;
        private static char[] IdentifierSeparator = new char[] { '[', ']', '(', ')', ':', '>',
            '<', '/', '+', '-', '*', '%', ',', '.', '=', ' ', '\t', '\n', '\r',
            '\u2212', '\u221e', '\u2215', '\u2217', '\u2227', '\u2228', '\u2254', '\u2264',
            '\u2265', '\u2260', '\u00F7', '\u00D7', '\u00AC', '\u2308', '\u2309', '\u230A',
            '\u230B', '\u2194', '{', '}', '\u222a', '\u2229', '\u2208', '\u2209', '\u2283',
            '\u2282', '\u2286', '\u2287', '\\', '\u2216', '!'
        };

        public Tokenizer(TextReader reader)
        {
            this.Reader = new TokenizerReader(reader);
            IndentationLevels = new List<int>() { 0 };
            EmittedTokens = new List<Token>();
            CurrentIndentation = 0;
        }
        public IEnumerable<Token> Tokenize()
        {
            while (Reader.Peek() != -1)
            {
                EmittedTokens.Clear();
                TokenizeStatement();
                Emit(TokenType.EndStatement);
                //now we can iterate over and yield results
                foreach (Token token in EmittedTokens)
                {
                    yield return token;
                }
            }
            for (int i = 0; i < IndentationLevels.Count - 1; ++i)
            {
                yield return TokenType.EndBlock;
            }
            yield return TokenType.EndInput;
        }


        private string ReadIdentifier(out bool dotRead)
        {
            //#number.#number or #number{base} or letters
            //beware of #number.. or #number{base}.. or #number.#number..
            RemoveWhitespace();
            int delta = Reader.Peek();
            if (delta >= '0' && delta <= '9') { return ReadNumber(out dotRead); }
            //identifier (true)
            string identifier = "";
            dotRead = false;
            while (true)
            {
                int ch = Reader.Peek();
                if (ch == -1 || IdentifierSeparator.Contains((char)ch)) { return identifier; }
                identifier += (char)ch;
                Reader.Read();
            }
        }
        private string ReadNumber(out bool dotRead)
        {
            dotRead = false;
            string integerPart = ReadIntegerDigits();
            int delta = Reader.Peek();
            //se trova una base allora il numero è terminato
            if (delta == 'o' || delta == 'd' || delta == 'b')
            {
                Reader.Read();
                return integerPart + (char)delta;
            }
            if (delta == '.')
            {
                //può essere un numero decimale oppure un operatore
                Reader.Read();
                dotRead = true;
                string decimalPart = ReadIntegerDigits();
                if (decimalPart.Length == 0)
                {
                    if (Reader.Peek() == '.')
                    {
                        //{number}..
                        return integerPart;
                    }
                    throw new InvalidDataException(string.Format("A number cannot have a dot without decimal part on line {0}, column {1} ", Reader.Line, Reader.Column));
                }
                dotRead = false;
                //{number}.{number}
                delta = Reader.Peek();
                if (delta == '.')
                {
                    //{number}.{number}.
                    dotRead = true;
                    Reader.Read();
                    delta = Reader.Peek();
                    if (delta == '.')
                    {
                        //{number}.{number}..
                        return integerPart + '.' + decimalPart;
                    }
                    throw new InvalidDataException(string.Format("A decimal number cannot be followed by only a dot on line {0}, column {1}", Reader.Line, Reader.Column));
                }
                if (delta == -1 || IdentifierSeparator.Contains((char)delta))
                {
                    return integerPart + '.' + decimalPart;
                }
                throw new InvalidDataException(string.Format("A number cannot contain letters on line {0}, column {1}", Reader.Line, Reader.Column));
            }
            if (delta == -1 || IdentifierSeparator.Contains((char)delta))
            {
                return integerPart;
            }
            throw new InvalidDataException(string.Format("A number cannot contain letters on line {0}, column {1}", Reader.Line, Reader.Column));
        }
        /// <summary>
        /// Legge una sequenza di caratteri fino a quando sono compresi tra 0 e 9 inclusi
        /// </summary>
        /// <returns>La sequenza di caratteri letta, o una stringa vuota</returns>
        private string ReadIntegerDigits()
        {
            string identifier = string.Empty;
            while (true)
            {
                int ch = Reader.Peek();
                if (ch < '0' || ch > '9') { return identifier; }
                identifier += (char)ch;
                Reader.Read();
            }

        }

        private int RemoveWhitespace()
        {
            int wsCount = 0;
            while (true)
            {
                int ch = Reader.Peek();
                if (ch == -1 || !IsWhitespace((char)ch) || ch == '\n')
                {
                    return wsCount;
                }
                Reader.Read();
                if (ch == '\t') { wsCount = wsCount + 8 - (wsCount % 8); }
                else { ++wsCount; }
            }
        }
        /// <summary>
        /// Indica se il carattere è classificato come spazio bianco
        /// </summary>
        /// <param name="c">Carattere da controllare</param>
        /// <returns>true se è uno spazio bianco, false altrimenti</returns>
        private static bool IsWhitespace(char c)
        {
            return c == ' ' || c == '\t' || c == '\n' || c == '\r';
        }

        /// <summary>
        /// Emette il token ed esegue alcuni controlli sull'indentazione e sul tipo di token
        /// </summary>
        /// <param name="token">Token da emettere</param>
        private void Emit(Token token)
        {
            //Imposta l'indentazione
            if (CurrentIndentation != -1 && token.TokenType != TokenType.EndInput)
            {
                if (token.TokenType == TokenType.EndStatement)
                {
                    //it's an empty statement
                    return;
                }
                int previousIndentation = IndentationLevels[IndentationLevels.Count - 1];
                if (previousIndentation < CurrentIndentation)
                {
                    //indent
                    IndentationLevels.Add(CurrentIndentation);
                    EmittedTokens.Add(TokenType.StartBlock);
                }
                else if (previousIndentation > CurrentIndentation)
                {
                    while (IndentationLevels.Count != 0 &&
                        (previousIndentation = IndentationLevels[IndentationLevels.Count - 1]) > CurrentIndentation)
                    {
                        //dedent
                        IndentationLevels.RemoveAt(IndentationLevels.Count - 1);
                        EmittedTokens.Add(TokenType.EndBlock);
                    }
                    if (previousIndentation != CurrentIndentation)
                    {
                        //ERROR
                        throw new InvalidDataException("Indentation level mismatch on line " + Reader.Line);
                    }
                }
                CurrentIndentation = -1;
            }
            EmittedTokens.Add(token);
        }
        
        /// <summary>
        /// Emette un token per un numero intero
        /// </summary>
        /// <param name="str">Stringa da convertire in token numerico</param>
        /// <returns>true se la conversione ha avuto successo, false altrimenti</returns>
        private bool TryEmitInteger(string str)
        {
            if (str.Contains('.')) { return false; }
            if (str[0] < '0' || str[0] > '9') { return false; }
            char cBase = str[str.Length - 1];
            int length = str.Length;
            int power = 0;
            switch (cBase)
            {
                case 'o':
                    //octal
                    --length;
                    power = 8;
                    break;
                case 'b':
                    //binary
                    --length;
                    power = 2;
                    break;
                case 'd':
                    //decimal with suffix
                    --length;
                    power = 10;
                    break;
                default:
                    //decimal
                    power = 10;
                    break;
            }
            BigInteger result = BigInteger.Zero;
            for (int i = 0; i < length; ++i)
            {
                if (str[i] > '0' + power - 1)
                {
                    //ERROR
                    throw new InvalidDataException(string.Format("Integer number is not on the specified base {0}", str));
                }
                result = result * power + (str[i] - '0');
            }
            Emit(new IntegerToken(result));
            return true;
        }

        /// <summary>
        /// Emette un token per un numero decimale
        /// </summary>
        /// <param name="str">Stringa da convertire in token numerico</param>
        /// <returns>true se la conversione ha avuto successo, false altrimenti</returns>
        private bool TryEmitFloat(string str)
        {
            if (str[0] < '0' || str[0] > '9') { return false; }
            double value = double.Parse(str, System.Globalization.CultureInfo.InvariantCulture);
            Emit(new FloatToken(value));
            return true;
        }

        /// <summary>
        /// Converte uno statement in una sequenza di token, in caso di errore, genera una eccezione
        /// </summary>
        private void TokenizeStatement()
        {
            //ottiene l'indentazione della linea corrente
            this.CurrentIndentation = RemoveWhitespace();
            //indica se l'ultimo token è stato un operatore
            bool wasOperator = false;
            //indica il numero di parentesi ( ancora aperte
            int groupOpened = 0;
            //indica il numero di parentesi [ ancora aperte
            int arrayOpened = 0;
            //indica il numero di parte intera superiore sinistre
            int ceilingOpened = 0;
            //indica il numero di parte intera inferiore sinistre
            int floorOpened = 0;
            //indica il numero di parentesi { ancora aperte
            int setOpened = 0;
            //indica se nel leggere un identificatore è stato letto anche un punto terminale(questo accade solo per i numeri)
            bool dotRead;
            //NB l'apertura di parentesi per convenzione comporta l'impostazione di wasOperator a false 

            while (true)
            {
                //identificatore letto
                string identifier = ReadIdentifier(out dotRead);
                if (identifier.Length == 0)
                {
                    //se non è stato letto un identificatore valido
                    //leggi il primo carattere
                    int ch = Reader.Peek();
                    switch (ch)
                    {
                        case -1: //fine dell'input
                            //se è rimasta aperta una parentesi o l'ultimo token è stato un operatore allora genera un errore
                            //altrimenti la procedura è terminata
                            if (arrayOpened > 0)
                            {
                                throw new InvalidDataException(string.Format("Unmatched '[' found on line {0}, column {1}", Reader.Line, Reader.Column));
                            }
                            if (groupOpened > 0)
                            {
                                throw new InvalidDataException(string.Format("Unmatched '(' found on line {0}, column {1}", Reader.Line, Reader.Column));
                            }
                            if (setOpened > 0)
                            {
                                throw new InvalidDataException(string.Format("Unmatched '{{' found on line {0}, column {1}", Reader.Line, Reader.Column));
                            }
                            if (ceilingOpened > 0)
                            {
                                throw new InvalidDataException(string.Format("Unmatched '\u2308' found on line {0}, column {1}", Reader.Line, Reader.Column));
                            }
                            if (floorOpened > 0)
                            {
                                throw new InvalidDataException(string.Format("Unmatched '\u230A' found on line {0}, column {1}", Reader.Line, Reader.Column));
                            }
                            if (wasOperator)
                            {
                                throw new InvalidDataException(string.Format("No argument found on line {0}, column {1}", Reader.Line, Reader.Column));
                            }
                            return;
                        case '\n': //fine della riga
                            //consuma il carattere, se vi sono ancora parentesi aperte o l'ultimo token è stato un operatore
                            //allora lo statement non è terminato, semplicemente continua con la linea successiva
                            Reader.Read();
                            if (arrayOpened > 0 || groupOpened > 0 || floorOpened > 0 || ceilingOpened > 0 || wasOperator)
                            {
                                //lo statement non è terminato
                                break;
                            }
                            return;
                        case '/':
                            //può essere un simbolo di divisone(/) oppure di commento(//)
                            Reader.Read();
                            if (Reader.Peek() == '/')
                            {
                                //è un commento, quindi consuma tutti i caratteri fino alla fine dello statement o dell'input (esclusi)
                                while (true)
                                {
                                    Reader.Read();
                                    ch = Reader.Peek();
                                    if (ch == '\n' || ch == -1) { break; }
                                }
                            }
                            else
                            {
                                //è una divisione
                                Emit(TokenType.Div);
                                wasOperator = true;
                                RemoveWhitespace();
                            }
                            break;
                        case '<':
                            //può essere minore(<), minore o uguale (<=), diverso (<>) ed anche swap(<->)
                            Reader.Read();
                            ch = Reader.Peek();
                            wasOperator = true;

                            if (ch == '=') { Emit(TokenType.LessEqual); Reader.Read(); }
                            else if (ch == '>') { Emit(TokenType.Different); Reader.Read(); }
                            else if (ch == '-')
                            {
                                //abbiamo letto <-
                                //se il carattere successivo è > allora è swap, altrimenti sono i due operatori < e -
                                Reader.Read();
                                ch = Reader.Peek();
                                if (ch == '>')
                                {
                                    Emit(TokenType.DoubleArrow);
                                    Reader.Read();
                                }
                                else
                                {
                                    Emit(TokenType.Less);
                                    Emit(TokenType.Minus);
                                }
                            }
                            else { Emit(TokenType.Less); }
                            RemoveWhitespace();
                            break;
                        case '\u2264':
                            //alias di <=, simbolo unicode
                            wasOperator = true;
                            Reader.Read();
                            Emit(TokenType.LessEqual);
                            RemoveWhitespace();
                            break;
                        case '\u2260':
                            //alias di <>, simbolo unicode
                            wasOperator = true;
                            Reader.Read();
                            Emit(TokenType.Different);
                            RemoveWhitespace();
                            break;
                        case '>':
                            //può essere maggiore(>) oppure maggiore o uguale(>=)
                            Reader.Read();
                            ch = Reader.Peek();
                            wasOperator = true;

                            if (ch == '=') { Emit(TokenType.GreaterEqual); Reader.Read(); }
                            else { Emit(TokenType.Greater); }
                            RemoveWhitespace();
                            break;
                        case '\u2265':
                            //alias di >=, simbolo unicode
                            wasOperator = true;
                            Reader.Read();
                            Emit(TokenType.GreaterEqual);
                            RemoveWhitespace();
                            break;
                        case '.':
                            //può essere l'operatore range(..) o l'accessore di campo (.)
                            Reader.Read();
                            ch = Reader.Peek();
                            wasOperator = true;

                            if (ch == '.') { Emit(TokenType.Range); Reader.Read(); }
                            else { Emit(TokenType.MemberAccessor); }
                            RemoveWhitespace();
                            break;
                        case '(':
                            wasOperator = false;
                            ++groupOpened;
                            Reader.Read();
                            Emit(TokenType.StartGroup);
                            RemoveWhitespace();
                            break;
                        case ')':
                            --groupOpened;
                            if (groupOpened < 0)
                            {
                                throw new InvalidDataException(string.Format("Unmatched ')' found on line {0}, column {1}", Reader.Line, Reader.Column));
                            }

                            Reader.Read();
                            Emit(TokenType.EndGroup);
                            RemoveWhitespace();
                            break;
                        case '[':
                            wasOperator = false;
                            ++arrayOpened;
                            Reader.Read();
                            Emit(TokenType.StartIndexer);
                            RemoveWhitespace();
                            break;
                        case ']':
                            --arrayOpened;
                            if (arrayOpened < 0)
                            {
                                throw new InvalidDataException(string.Format("Unmatched ']' found on line {0}, column {1}", Reader.Line, Reader.Column));
                            }

                            Reader.Read();
                            Emit(TokenType.EndIndexer);
                            RemoveWhitespace();
                            break;
                        case '{':
                            wasOperator = false;
                            Reader.Read();
                            RemoveWhitespace();
                            if (Reader.Peek() == '}')
                            {
                                //E' stato letto {}
                                Emit(TokenType.EmptySet);
                                Reader.Read();
                                RemoveWhitespace();
                            }
                            else
                            {
                                Emit(TokenType.StartSet);
                                ++setOpened;
                            }
                            break;
                        case '}':
                            --setOpened;
                            if (setOpened < 0)
                            {
                                throw new InvalidDataException(string.Format("Unmatched '}}' found on line {0}, column {1}", Reader.Line, Reader.Column));
                            }

                            Reader.Read();
                            Emit(TokenType.EndSet);
                            RemoveWhitespace();
                            break;
                        case '\u2308'://parte intera superiore sinistra
                            wasOperator = false;
                            ++ceilingOpened;
                            Reader.Read();
                            Emit(TokenType.StartCeiling);
                            RemoveWhitespace();
                            break;
                        case '\u2309'://parte interna superiore destra
                            --ceilingOpened;
                            if (ceilingOpened < 0)
                            {
                                throw new InvalidDataException(string.Format("Unmatched '\u2309' found on line {0}, column {1}", Reader.Line, Reader.Column));
                            }

                            Reader.Read();
                            Emit(TokenType.EndCeiling);
                            RemoveWhitespace();
                            break;
                        case '\u230A'://parte intera inferiore sinistra, unicode
                            wasOperator = false;
                            ++floorOpened;
                            Reader.Read();
                            Emit(TokenType.StartFloor);
                            RemoveWhitespace();
                            break;
                        case '\u230B'://parte intera inferiore destra, unicode
                            --floorOpened;
                            if (floorOpened < 0)
                            {
                                throw new InvalidDataException(string.Format("Unmatched '\u230B' found on line {0}, column {1}", Reader.Line, Reader.Column));
                            }

                            Reader.Read();
                            Emit(TokenType.EndFloor);
                            RemoveWhitespace();
                            break;
                        case ',':
                            wasOperator = true;
                            Reader.Read();
                            Emit(TokenType.EndArgument);
                            RemoveWhitespace();
                            break;
                        case '\u2194':
                            //alias di <->, simbolo unicode
                            wasOperator = true;
                            Reader.Read();
                            Emit(TokenType.DoubleArrow);
                            RemoveWhitespace();
                            break;
                        case '+':
                            wasOperator = true;
                            Reader.Read();
                            Emit(TokenType.Plus);
                            RemoveWhitespace();
                            break;
                        case '\u2215'://division slash
                        case '\u00F7'://:- sign (obelo)
                            wasOperator = true;
                            Reader.Read();
                            Emit(TokenType.Div);
                            RemoveWhitespace();
                            break;
                        case '-':
                        case '\u2212'://simbolo unicode -
                            wasOperator = true;
                            Reader.Read();
                            Emit(TokenType.Minus);
                            RemoveWhitespace();
                            break;
                        case '*':
                        case '\u2217'://asterisk multiplication sign unicode
                        case '\u00D7'://multiplication x sign unicode
                            wasOperator = true;
                            Reader.Read();
                            Emit(TokenType.Mult);
                            RemoveWhitespace();
                            break;
                        case '%':
                            wasOperator = true;
                            Reader.Read();
                            Emit(TokenType.Remainder);
                            RemoveWhitespace();
                            break;
                        case ':':
                            Reader.Read();
                            ch = Reader.Read();
                            if (ch == '=')
                            {
                                wasOperator = true;
                                Emit(TokenType.Assignment);
                                RemoveWhitespace();
                            }
                            else
                            {
                                throw new InvalidDataException(string.Format("Found '{0}', expected '=' on line {1}, column {2}", ch==-1?"EOI":((char)ch).ToString(), Reader.Line, Reader.Column));
                            }
                            break;
                        case '\u2254'://unicode assegnamento( := )
                            wasOperator = true;
                            Reader.Read();
                            Emit(TokenType.Assignment);
                            RemoveWhitespace();
                            break;
                        case '=':
                            wasOperator = true;
                            Reader.Read();
                            Emit(TokenType.Equal);
                            RemoveWhitespace();
                            break;
                        case '\u221e'://unicode infinito
                            wasOperator = false;
                            Reader.Read();
                            Emit(TokenType.Infinity);
                            RemoveWhitespace();
                            break;
                        case '\u2227'://unicode and logico
                            wasOperator = true;
                            Reader.Read();
                            Emit(TokenType.LogicalAnd);
                            RemoveWhitespace();
                            break;
                        case '\u2228'://unicode or logico
                            wasOperator = true;
                            Reader.Read();
                            Emit(TokenType.LogicalOr);
                            RemoveWhitespace();
                            break;
                        case '!':
                            Reader.Read();
                            ch = Reader.Peek();
                            if (ch == '=')
                            {
                                wasOperator = true;
                                Reader.Read();
                                Emit(TokenType.Different);
                                RemoveWhitespace();
                            }
                            else
                            {
                                throw new InvalidDataException(string.Format("Found '{0}', expected '=' on line {1}, column {2}", ch == -1 ? "EOI" : ((char)ch).ToString(), Reader.Line, Reader.Column));
                            }
                            break;
                        case '\u00AC'://unicode not logico
                            wasOperator = true;
                            Reader.Read();
                            Emit(TokenType.LogicalNot);
                            RemoveWhitespace();
                            break;
                        case '\u2282'://unicode subset <
                            wasOperator = true;
                            Reader.Read();
                            Emit(TokenType.SubsetOf);
                            RemoveWhitespace();
                            break;
                        case '\u2283'://unicode superset >
                            wasOperator = true;
                            Reader.Read();
                            Emit(TokenType.SupersetOf);
                            RemoveWhitespace();
                            break;
                        case '\u2286'://unicode subsetEqual <=
                            wasOperator = true;
                            Reader.Read();
                            Emit(TokenType.SubsetOfEqual);
                            RemoveWhitespace();
                            break;
                        case '\u2287'://unicode supersetEqual >=
                            wasOperator = true;
                            Reader.Read();
                            Emit(TokenType.SupersetOfEqual);
                            RemoveWhitespace();
                            break;
                        case '\u222A'://unicode union U
                            wasOperator = true;
                            Reader.Read();
                            Emit(TokenType.Union);
                            RemoveWhitespace();
                            break;
                        case '\u2229'://unicode intersection Ureverse
                            wasOperator = true;
                            Reader.Read();
                            Emit(TokenType.Intersection);
                            RemoveWhitespace();
                            break;
                        case '\u2208'://unicode elementOf
                            wasOperator = true;
                            Reader.Read();
                            Emit(TokenType.ElementOf);
                            RemoveWhitespace();
                            break;
                        case '\u2209'://unicode not elementOf
                            wasOperator = true;
                            Reader.Read();
                            Emit(TokenType.NotElementOf);
                            RemoveWhitespace();
                            break;
                        case '\u2216'://unicode setMinus
                        case '\\':
                            wasOperator = true;
                            Reader.Read();
                            Emit(TokenType.SetMinus);
                            RemoveWhitespace();
                            break;
                        default:
                            //ERROR
                            throw new InvalidDataException(string.Format("Found '{0}', on line {1}, column {2}", ch == -1 ? "EOI" : ((char)ch).ToString(), Reader.Line, Reader.Column));
                    }
                }
                else
                {
                    wasOperator = false;
                    //può essere un identificatore, un numero, una parola chiave, una costante(nil, infinity, {emptySet})
                    switch (identifier)
                    {
                        case "\u2205": Emit(TokenType.EmptySet); break;
                        case "for": Emit(TokenType.For); break;
                        case "else": Emit(TokenType.Else); break;
                        case "if": Emit(TokenType.If); break;
                        case "while": Emit(TokenType.While); break;
                        case "return": Emit(TokenType.Return); break;
                        case "to": Emit(TokenType.To); break;
                        case "and": Emit(TokenType.LogicalAnd); break;
                        case "or": Emit(TokenType.LogicalOr); break;
                        case "not": Emit(TokenType.LogicalNot); break;
                        case "infinity": Emit(TokenType.Infinity); break;
                        case "nil": Emit(TokenType.Nil); break;
                        case "repeat": Emit(TokenType.Repeat); break;
                        case "until": Emit(TokenType.Until); break;
                        case "downto": Emit(TokenType.Downto); break;
                        case "swap": Emit(TokenType.Swap); break;
                        case "each": Emit(TokenType.Each); break;
                        case "then": Emit(TokenType.Then); break;
                        case "foreach":
                            Emit(TokenType.For);
                            Emit(TokenType.Each);
                            break;
                        case "print": Emit(TokenType.Print); break;
                        case "elsif":
                            Emit(TokenType.Else);
                            Emit(TokenType.If);
                            break;
                        default:
                            //non è una keyword, può essere un intero, un decimale o una stringa
                            if (!TryEmitInteger(identifier) && !TryEmitFloat(identifier))
                            {
                                Emit(new IdentifierToken(identifier));
                            }
                            break;
                    }
                    if (dotRead)
                    {
                        //se nell'identificatore è stato letto un punto
                        //bisogna verificare il carattere successivo
                        //ed emettere al limite un token
                        //Questo è necessario per permettere scritture di questo tipo:  A[1..2]
                        int ch = Reader.Peek();
                        if (ch == '.')
                        {
                            Emit(TokenType.Range);
                            Reader.Read();
                            RemoveWhitespace();
                        }
                        else
                        {
                            Emit(TokenType.MemberAccessor);
                            RemoveWhitespace();
                        }
                        wasOperator = true;
                    }
                }
            }
        }
    }
}
